# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import asyncer
import dagger
import toml
from consts import (
    DOCKER_HOST_NAME,
    DOCKER_HOST_PORT,
    DOCKER_VERSION,
    POETRY_CACHE_VOLUME_NAME,
    PYPROJECT_TOML_FILE_PATH,
)
from pipelines.airbyte_ci.test.models import deserialize_airbyte_ci_config
from pipelines.dagger.actions.system import docker
from pipelines.helpers.github import update_commit_status_check
from pipelines.helpers.utils import sh_dash_c
from pipelines.models.steps import PoeTaskResult, StepStatus

if TYPE_CHECKING:
    from logging import Logger
    from pathlib import Path
    from typing import Dict, List

    from pipelines.airbyte_ci.test.models import AirbyteCiPackageConfiguration

# The following directories are always mounted because a lot of tests rely on them
DIRECTORIES_TO_ALWAYS_MOUNT = [
    ".git",  # This is needed as some package tests rely on being in a git repo
    ".github",
    "docs",
    "airbyte-integrations",
    "airbyte-ci",
    "airbyte-cdk",
    PYPROJECT_TOML_FILE_PATH,
    "LICENSE_SHORT",
    "poetry.lock",
    "spotless-maven-pom.xml",
    "tools/gradle/codestyle/java-google-style.xml",
]

DEFAULT_EXCLUDE = [
    "**/__pycache__",
    "**/.pytest_cache",
    "**/.venv",
    "**.log",
    "**/.gradle",
]

DEFAULT_CONTAINER_IMAGE = "python:{version}"

VERSION_CONTAINER_IMAGES = {
    "3.10": DEFAULT_CONTAINER_IMAGE.format(version="3.10.12"),
    "3.11": DEFAULT_CONTAINER_IMAGE.format(version="3.11.9"),
}


async def get_filtered_airbyte_repo_dir(
    dagger_client: dagger.Client, poetry_package_path: Path
) -> dagger.Directory:
    """Get a filtered airbyte repo directory with the directories to always mount and the poetry package path.

    Args:
        dagger_client (dagger.Client): Dagger client.
        poetry_package_path (Path): Path to the poetry package in the airbyte repo.

    Returns:
        dagger.Directory: The filtered airbyte repo directory.
    """
    directories_to_mount = list(
        set([str(poetry_package_path), *DIRECTORIES_TO_ALWAYS_MOUNT])
    )
    return dagger_client.host().directory(
        ".",
        exclude=DEFAULT_EXCLUDE,
        include=directories_to_mount,
    )


async def get_poetry_package_dir(
    airbyte_repo_dir: dagger.Directory, poetry_package_path: Path
) -> dagger.Directory:
    """Get the poetry package directory from the airbyte repo directory.

    Args:
        airbyte_repo_dir (dagger.Directory): The airbyte repo directory.
        poetry_package_path (Path): Path to the poetry package in the airbyte repo.

    Raises:
        FileNotFoundError: If the pyproject.toml file is not found in the poetry package directory.
        FileNotFoundError: If the poetry package directory is not found in the airbyte repo directory.

    Returns:
        dagger.Directory: The poetry package directory.
    """
    try:
        package_directory = await airbyte_repo_dir.directory(str(poetry_package_path))
        if PYPROJECT_TOML_FILE_PATH not in await package_directory.entries():
            raise FileNotFoundError(
                f"Could not find pyproject.toml in {poetry_package_path}, are you sure this is a poetry package?"
            )
    except dagger.DaggerError:
        raise FileNotFoundError(
            f"Could not find {poetry_package_path} in the repository, are you sure this path is correct?"
        )
    return package_directory


async def get_airbyte_ci_package_config(
    poetry_package_dir: dagger.Directory,
) -> AirbyteCiPackageConfiguration:
    """Get the airbyte ci package configuration from the pyproject.toml file in the poetry package directory.

    Args:
        poetry_package_dir (dagger.Directory): The poetry package directory.

    Returns:
        AirbyteCiPackageConfiguration: The airbyte ci package configuration.
    """
    raw_pyproject_toml = await poetry_package_dir.file(
        PYPROJECT_TOML_FILE_PATH
    ).contents()
    pyproject_toml = toml.loads(raw_pyproject_toml)
    return deserialize_airbyte_ci_config(pyproject_toml)


def get_poetry_base_container(
    dagger_client: dagger.Client, python_version: str
) -> dagger.Container:
    """Get a base container with system dependencies to run poe tasks of poetry package:
    - git: required for packages using GitPython
    - poetry
    - poethepoet
    - docker: required for packages using docker in their tests

    Args:
        dagger_client (dagger.Client): The dagger client.

    Returns:
        dagger.Container: The base container.
    """
    poetry_cache_volume: dagger.CacheVolume = dagger_client.cache_volume(
        POETRY_CACHE_VOLUME_NAME
    )
    poetry_cache_path = "/root/.cache/poetry"
    container_image = VERSION_CONTAINER_IMAGES.get(
        python_version, DEFAULT_CONTAINER_IMAGE.format(version=python_version)
    )
    return (
        dagger_client.container()
        .from_(container_image)
        .with_env_variable("PIPX_BIN_DIR", "/usr/local/bin")
        .with_env_variable("POETRY_CACHE_DIR", poetry_cache_path)
        .with_mounted_cache(poetry_cache_path, poetry_cache_volume)
        .with_exec(
            sh_dash_c(
                [
                    "apt-get update",
                    "apt-get install -y bash git curl",
                    "pip install pipx",
                    "pipx ensurepath",
                    "pipx install poetry",
                    "pipx install poethepoet",
                ]
            )
        )
        .with_env_variable("VERSION", DOCKER_VERSION)
        .with_exec(sh_dash_c(["curl -fsSL https://get.docker.com | sh"]))
    )


def prepare_container_for_poe_tasks(
    dagger_client: dagger.Client,
    airbyte_repo_dir: dagger.Directory,
    airbyte_ci_package_config: AirbyteCiPackageConfiguration,
    poetry_package_path: Path,
    pipeline_context_params: Dict,
    python_version: str,
) -> dagger.Container:
    """Prepare a container to run poe tasks for a poetry package.

    Args:
        dagger_client (dagger.Client): The dagger client.
        airbyte_repo_dir (dagger.Directory): The airbyte repo directory.
        airbyte_ci_package_config (AirbyteCiPackageConfiguration): The airbyte ci package configuration.
        poetry_package_path (Path): The path to the poetry package in the airbyte repo.
        pipeline_context_params (Dict): The pipeline context parameters.

    Returns:
        dagger.Container: The container to run poe tasks for the poetry package.
    """

    # BE CAREFUL ABOUT THE ORDER OF THESE INSTRUCTIONS
    # PLEASE REMIND THAT DAGGER OPERATION ARE CACHED LIKE IN DOCKERFILE:
    # ANY CHANGE IN THE INPUTS OF AN OPERATION WILL INVALIDATE THE DOWNSTREAM OPERATIONS CACHE

    # Start from the base container
    container = get_poetry_base_container(dagger_client, python_version)

    # Set the CI environment variable
    is_ci = pipeline_context_params["is_ci"]
    if is_ci:
        container = container.with_env_variable("CI", "true")

    # Bind to dockerd service if needed
    if airbyte_ci_package_config.side_car_docker_engine:
        dockerd_service = docker.with_global_dockerd_service(dagger_client)
        container = (
            container.with_env_variable(
                "DOCKER_HOST", f"tcp://{DOCKER_HOST_NAME}:{DOCKER_HOST_PORT}"
            )
            .with_env_variable("DOCKER_HOST_NAME", DOCKER_HOST_NAME)
            .with_service_binding(DOCKER_HOST_NAME, dockerd_service)
        )

    # Mount the docker socket if needed
    if airbyte_ci_package_config.mount_docker_socket:
        container = container.with_unix_socket(
            "/var/run/docker.sock",
            dagger_client.host().unix_socket("/var/run/docker.sock"),
        )

    # Set the required environment variables according to the package configuration
    for required_env_var in airbyte_ci_package_config.required_environment_variables:
        # We consider any environment variable as a secret for safety reasons
        secret_env_var = dagger_client.set_secret(
            required_env_var, os.environ[required_env_var]
        )
        container = container.with_secret_variable(required_env_var, secret_env_var)

    # Mount the airbyte repo directory
    container = container.with_mounted_directory("/airbyte", airbyte_repo_dir)

    # Set working directory to the poetry package directory
    container = container.with_workdir(f"/airbyte/{poetry_package_path}")

    # If a package from `airbyte-platform-internal` is required, modify the entry in pyproject.toml to use https instead of ssh,
    # when run in Github Actions
    # This is currently required for getting the connection-retriever package, for regression tests.
    if is_ci:
        container = (
            container.with_exec(
                [
                    "sed",
                    "-i",
                    "-E",
                    r"s,git@github\.com:airbytehq/airbyte-platform-internal,https://github.com/airbytehq/airbyte-platform-internal.git,",
                    "pyproject.toml",
                ],
                use_entrypoint=True,
            )
            .with_exec(
                [
                    "poetry",
                    "source",
                    "add",
                    "--priority=supplemental",
                    "airbyte-platform-internal-source",
                    "https://github.com/airbytehq/airbyte-platform-internal.git",
                ],
                use_entrypoint=True,
            )
            .with_secret_variable(
                "CI_GITHUB_ACCESS_TOKEN",
                dagger_client.set_secret(
                    "CI_GITHUB_ACCESS_TOKEN",
                    pipeline_context_params["ci_github_access_token"].value,
                ),
            )
            .with_exec(
                [
                    "/bin/sh",
                    "-c",
                    "poetry config http-basic.airbyte-platform-internal-source octavia-squidington-iii $CI_GITHUB_ACCESS_TOKEN",
                ],
                use_entrypoint=True,
            )
            .with_exec(["poetry", "lock"], use_entrypoint=True)
        )

    # Install the poetry package
    container = container.with_exec(
        ["poetry", "install"]
        + [
            f"--with={group}"
            for group in airbyte_ci_package_config.optional_poetry_groups
        ]
        + [f"--extras={extra}" for extra in airbyte_ci_package_config.poetry_extras]
    )
    return container


async def run_poe_task(container: dagger.Container, poe_task: str) -> PoeTaskResult:
    """Run the poe task in the container and return a PoeTaskResult.

    Args:
        container (dagger.Container): The container to run the poe task in.
        poe_task (str): The poe task to run.

    Returns:
        PoeTaskResult: The result of the command execution.
    """
    try:
        executed_container = await container.with_exec(
            ["poe", poe_task], use_entrypoint=True
        )
        return PoeTaskResult(
            task_name=poe_task,
            status=StepStatus.SUCCESS,
            stdout=await executed_container.stdout(),
            stderr=await executed_container.stderr(),
        )
    except dagger.ExecError as e:
        return PoeTaskResult(task_name=poe_task, status=StepStatus.FAILURE, exc_info=e)


async def run_and_log_poe_task_results(
    pipeline_context_params: Dict,
    package_name: str,
    container: dagger.Container,
    poe_task: str,
    logger: Logger,
) -> PoeTaskResult:
    """Run the poe task in the container and log the result.

    Args:
        pipeline_context_params (Dict): The pipeline context parameters.
        package_name (str): The name of the package to run the poe task for.
        container (dagger.Container): The container to run the poe task in.
        poe_task (str): The poe task to run.
        logger (Logger): The logger to log the result.

    Returns:
        PoeTaskResult: The result of the command execution.
    """

    commit_status_check_params = {
        "sha": pipeline_context_params["git_revision"],
        "description": f"{poe_task} execution for {package_name}",
        "context": f"{package_name} - {poe_task}",
        "target_url": f"{pipeline_context_params['gha_workflow_run_url']}",
        "should_send": pipeline_context_params["is_ci"],
        "logger": logger,
    }

    logger.info(f"Running poe task: {poe_task}")
    # Send pending status check
    update_commit_status_check(**{**commit_status_check_params, "state": "pending"})
    result = await run_poe_task(container, poe_task)
    result.log(logger)
    # Send the final status check
    update_commit_status_check(
        **{**commit_status_check_params, "state": result.status.get_github_state()}
    )

    return result


async def run_poe_tasks_for_package(
    dagger_client: dagger.Client,
    poetry_package_path: Path,
    pipeline_context_params: Dict,
) -> List[PoeTaskResult]:
    """Concurrently Run the poe tasks declared in pyproject.toml for a poetry package.

    Args:
        dagger_client (dagger.Client): The dagger client.
        poetry_package_path (Path): The path to the poetry package in the airbyte repo.
        pipeline_context_params (Dict): The pipeline context parameters.
    Returns:
        List[PoeTaskResult]: The results of the poe tasks.
    """
    dagger_client = dagger_client
    airbyte_repo_dir = await get_filtered_airbyte_repo_dir(
        dagger_client, poetry_package_path
    )
    package_dir = await get_poetry_package_dir(airbyte_repo_dir, poetry_package_path)
    package_config = await get_airbyte_ci_package_config(package_dir)

    logger = logging.getLogger(str(poetry_package_path))

    if not package_config.poe_tasks:
        logger.warning("No poe tasks to run.")
        return []

    logger.info(f"Python versions: {package_config.python_versions}")

    poe_task_results: List[asyncer.SoonValue] = []
    return_results = []

    for python_version in package_config.python_versions:
        container = prepare_container_for_poe_tasks(
            dagger_client,
            airbyte_repo_dir,
            package_config,
            poetry_package_path,
            pipeline_context_params,
            python_version,
        )

        async with asyncer.create_task_group() as poe_tasks_task_group:
            for task in package_config.poe_tasks:
                poe_task_results.append(
                    poe_tasks_task_group.soonify(run_and_log_poe_task_results)(
                        pipeline_context_params,
                        str(poetry_package_path),
                        container,
                        task,
                        logger.getChild(f"@{python_version}"),
                    )
                )

        return_results.extend([result.value for result in poe_task_results])

    return return_results



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""SQL query templates and schema documentation for Airbyte Cloud Prod DB Replica.

Prod DB Replica Schema Reference
================================

Database: prod-configapi
Instance: prod-ab-cloud-proj:us-west3:prod-pgsql-replica

connection
----------
id, namespace_definition, namespace_format, prefix, source_id, destination_id, name,
catalog, status, schedule, manual, resource_requirements, created_at, updated_at,
source_catalog_id, schedule_type, schedule_data, non_breaking_change_preference,
breaking_change, field_selection_data, destination_catalog_id, status_reason

actor
-----
id, workspace_id, actor_definition_id, name, configuration, actor_type, tombstone,
created_at, updated_at, resource_requirements

workspace
---------
id, customer_id, name, slug, email, initial_setup_complete, anonymous_data_collection,
send_newsletter, send_security_updates, display_setup_wizard, tombstone, notifications,
first_sync_complete, feedback_complete, created_at, updated_at, webhook_operation_configs,
notification_settings, organization_id, dataplane_group_id

dataplane_group
---------------
id, organization_id, name, enabled, created_at, updated_at, tombstone

Note: Main dataplane groups are:
- 645a183f-b12b-4c6e-8ad3-99e165603450 = US (default, ~133K workspaces)
- 153996d3-208e-4887-b8b1-e5fe48104450 = US-Central (~12K workspaces)
- b9e48d61-f082-4a14-a8d0-799a907938cb = EU (~3K workspaces)

actor_definition_version
------------------------
id, actor_definition_id, created_at, updated_at, documentation_url, docker_repository,
docker_image_tag, spec, protocol_version, release_date, normalization_repository,
normalization_tag, supports_dbt, normalization_integration_type, allowed_hosts,
suggested_streams, release_stage, support_state, support_level, supports_refreshes,
cdk_version, last_published, internal_support_level, language, supports_file_transfer,
supports_data_activation, connector_ipc_options

scoped_configuration
--------------------
id, key, resource_type, resource_id, scope_type, scope_id, value, description,
reference_url, origin_type, origin, expires_at, created_at, updated_at

Note: Version overrides are stored with key='connector_version', resource_type='actor_definition',
scope_type='actor', and value=actor_definition_version.id (UUID).

jobs
----
id, config_type, scope (connection_id), config, status, started_at, created_at,
updated_at, metadata, is_scheduled

Note: status values: 'succeeded', 'failed', 'cancelled', 'running', 'incomplete'
      config_type values: 'sync', 'reset_connection', 'refresh'

attempts
--------
id, job_id, attempt_number, log_path, output, status, created_at, updated_at,
ended_at, failure_summary, processing_task_queue, attempt_sync_config

connector_rollout
-----------------
id, actor_definition_id, release_candidate_version_id, initial_version_id, state,
initial_rollout_pct, current_target_rollout_pct, final_target_rollout_pct,
has_breaking_changes, max_step_wait_time_mins, updated_by, created_at, updated_at,
completed_at, expires_at, error_msg, failed_reason, rollout_strategy, workflow_run_id,
paused_reason, filters, tag

Note: state values: 'initialized', 'workflow_started', 'in_progress', 'paused',
      'finalizing', 'succeeded', 'errored', 'failed_rolled_back', 'canceled'
      Active states: initialized, workflow_started, in_progress, paused, finalizing, errored
      Terminal states: succeeded, failed_rolled_back, canceled
"""

from __future__ import annotations

import sqlalchemy

# =============================================================================
# Connection Queries
# =============================================================================

# Query connections by connector type (no organization filter)
# Note: pg8000 cannot determine the type of NULL parameters in patterns like
# "(:param IS NULL OR column = :param)", so we use separate queries instead
SELECT_CONNECTIONS_BY_CONNECTOR = sqlalchemy.text(
    """
    SELECT
         connection.id AS connection_id,
         connection.name AS connection_name,
         connection.source_id,
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name,
         source_actor.actor_definition_id AS source_definition_id,
         source_actor.name AS source_name
    FROM connection
    JOIN actor AS source_actor
      ON connection.source_id = source_actor.id
    JOIN workspace
      ON source_actor.workspace_id = workspace.id
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    WHERE
         source_actor.actor_definition_id = :connector_definition_id
    LIMIT :limit
    """
)

# Query connections by connector type, filtered by organization
SELECT_CONNECTIONS_BY_CONNECTOR_AND_ORG = sqlalchemy.text(
    """
    SELECT
         connection.id AS connection_id,
         connection.name AS connection_name,
         connection.source_id,
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name,
         source_actor.actor_definition_id AS source_definition_id,
         source_actor.name AS source_name
    FROM connection
    JOIN actor AS source_actor
      ON connection.source_id = source_actor.id
    JOIN workspace
      ON source_actor.workspace_id = workspace.id
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    WHERE
         source_actor.actor_definition_id = :connector_definition_id
     AND workspace.organization_id = :organization_id
    LIMIT :limit
    """
)

# Query connections by DESTINATION connector type (no organization filter)
SELECT_CONNECTIONS_BY_DESTINATION_CONNECTOR = sqlalchemy.text(
    """
    SELECT
         connection.id AS connection_id,
         connection.name AS connection_name,
         connection.destination_id,
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name,
         destination_actor.actor_definition_id AS destination_definition_id,
         destination_actor.name AS destination_name
    FROM connection
    JOIN actor AS destination_actor
      ON connection.destination_id = destination_actor.id
    JOIN workspace
      ON destination_actor.workspace_id = workspace.id
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    WHERE
         destination_actor.actor_definition_id = :connector_definition_id
    LIMIT :limit
    """
)

# Query connections by DESTINATION connector type, filtered by organization
SELECT_CONNECTIONS_BY_DESTINATION_CONNECTOR_AND_ORG = sqlalchemy.text(
    """
    SELECT
         connection.id AS connection_id,
         connection.name AS connection_name,
         connection.destination_id,
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name,
         destination_actor.actor_definition_id AS destination_definition_id,
         destination_actor.name AS destination_name
    FROM connection
    JOIN actor AS destination_actor
      ON connection.destination_id = destination_actor.id
    JOIN workspace
      ON destination_actor.workspace_id = workspace.id
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    WHERE
         destination_actor.actor_definition_id = :connector_definition_id
     AND workspace.organization_id = :organization_id
    LIMIT :limit
    """
)

# =============================================================================
# Connector Version Queries
# =============================================================================

SELECT_CONNECTOR_VERSIONS = sqlalchemy.text(
    """
    SELECT
         actor_definition_version.id AS version_id,
         actor_definition_version.docker_image_tag,
         actor_definition_version.docker_repository,
         actor_definition_version.release_stage,
         actor_definition_version.support_level,
         actor_definition_version.cdk_version,
         actor_definition_version.language,
         actor_definition_version.last_published,
         actor_definition_version.release_date
    FROM actor_definition_version
    WHERE
         actor_definition_version.actor_definition_id = :actor_definition_id
    ORDER BY
         actor_definition_version.last_published DESC NULLS LAST,
         actor_definition_version.created_at DESC
    """
)

# List new connector releases within the last N days
# Uses last_published (timestamp) rather than release_date (date only, often NULL)
# Note: No index on last_published, but table is small (~39K rows)
SELECT_NEW_CONNECTOR_RELEASES = sqlalchemy.text(
    """
    SELECT
         actor_definition_version.id AS version_id,
         actor_definition_version.actor_definition_id,
         actor_definition_version.docker_repository,
         actor_definition_version.docker_image_tag,
         actor_definition_version.last_published,
         actor_definition_version.release_date,
         actor_definition_version.release_stage,
         actor_definition_version.support_level,
         actor_definition_version.cdk_version,
         actor_definition_version.language,
         actor_definition_version.created_at
    FROM actor_definition_version
    WHERE
         actor_definition_version.last_published >= :cutoff_date
    ORDER BY
         actor_definition_version.last_published DESC
    LIMIT :limit
    """
)

SELECT_ACTORS_PINNED_TO_VERSION = sqlalchemy.text(
    """
    SELECT
         scoped_configuration.scope_id AS actor_id,
         scoped_configuration.resource_id AS actor_definition_id,
         scoped_configuration.origin_type,
         scoped_configuration.origin,
         scoped_configuration.description,
         scoped_configuration.created_at,
         scoped_configuration.expires_at,
         actor.name AS actor_name,
         actor.workspace_id,
         workspace.name AS workspace_name,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name
    FROM scoped_configuration
    JOIN actor
      ON scoped_configuration.scope_id = actor.id
    JOIN workspace
      ON actor.workspace_id = workspace.id
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    WHERE
         scoped_configuration.key = 'connector_version'
     AND scoped_configuration.scope_type = 'actor'
     AND scoped_configuration.value = :actor_definition_version_id
    ORDER BY
         scoped_configuration.created_at DESC
    """
)

# =============================================================================
# Sync Results Queries
# =============================================================================

# Get sync results for actors pinned to a specific connector definition VERSION ID
# This joins through scoped_configuration to find actors with version overrides
SELECT_SYNC_RESULTS_FOR_VERSION = sqlalchemy.text(
    """
    SELECT
         jobs.id AS job_id,
         jobs.scope AS connection_id,
         jobs.status AS job_status,
         jobs.started_at,
         jobs.updated_at AS job_updated_at,
         connection.name AS connection_name,
         actor.id AS actor_id,
         actor.name AS actor_name,
         actor.actor_definition_id,
         scoped_configuration.origin_type AS pin_origin_type,
         scoped_configuration.origin AS pin_origin,
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name
    FROM jobs
    JOIN connection
      ON jobs.scope = connection.id::text
    JOIN actor
      ON connection.source_id = actor.id
    JOIN scoped_configuration
      ON scoped_configuration.scope_id = actor.id
     AND scoped_configuration.key = 'connector_version'
     AND scoped_configuration.scope_type = 'actor'
    JOIN workspace
      ON actor.workspace_id = workspace.id
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    WHERE
         jobs.config_type = 'sync'
     AND scoped_configuration.value = :actor_definition_version_id
     AND jobs.updated_at >= :cutoff_date
    ORDER BY
         jobs.updated_at DESC
    LIMIT :limit
    """
)

# Get successful sync results for actors pinned to a specific connector definition VERSION ID
SELECT_SUCCESSFUL_SYNCS_FOR_VERSION = sqlalchemy.text(
    """
    SELECT
         jobs.id AS job_id,
         jobs.scope AS connection_id,
         jobs.started_at,
         jobs.updated_at AS job_updated_at,
         connection.name AS connection_name,
         actor.id AS actor_id,
         actor.name AS actor_name,
         actor.actor_definition_id,
         scoped_configuration.origin_type AS pin_origin_type,
         scoped_configuration.origin AS pin_origin,
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name
    FROM jobs
    JOIN connection
      ON jobs.scope = connection.id::text
    JOIN actor
      ON connection.source_id = actor.id
    JOIN scoped_configuration
      ON scoped_configuration.scope_id = actor.id
     AND scoped_configuration.key = 'connector_version'
     AND scoped_configuration.scope_type = 'actor'
    JOIN workspace
      ON actor.workspace_id = workspace.id
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    WHERE
         jobs.config_type = 'sync'
     AND jobs.status = 'succeeded'
     AND scoped_configuration.value = :actor_definition_version_id
     AND jobs.updated_at >= :cutoff_date
    ORDER BY
         jobs.updated_at DESC
    LIMIT :limit
    """
)

# Get recent sync results for ALL actors using a SOURCE connector definition.
# Finds all actors with the given actor_definition_id and returns their sync attempts,
# regardless of whether they have explicit version pins.
# Query starts from jobs table to leverage indexed columns.
# The LEFT JOIN to scoped_configuration provides pin context when available (pin_origin_type,
# pin_origin, pinned_version_id will be NULL for unpinned actors).
# Status filtering ('all', 'succeeded', 'failed') is handled at the application layer by
# selecting among different SQL query constants; this query returns all statuses.
SELECT_RECENT_SYNCS_FOR_SOURCE_CONNECTOR = sqlalchemy.text(
    """
    SELECT
         jobs.id AS job_id,
         jobs.scope AS connection_id,
         jobs.status AS job_status,
         jobs.started_at AS job_started_at,
         jobs.updated_at AS job_updated_at,
         connection.name AS connection_name,
         actor.id AS actor_id,
         actor.name AS actor_name,
         actor.actor_definition_id,
         actor.tombstone AS actor_tombstone,
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name,
         scoped_configuration.origin_type AS pin_origin_type,
         scoped_configuration.origin AS pin_origin,
         scoped_configuration.value AS pinned_version_id
    FROM jobs
    JOIN connection
      ON jobs.scope = connection.id::text
     AND connection.status != 'deprecated'
    JOIN actor
      ON connection.source_id = actor.id
     AND actor.actor_definition_id = :connector_definition_id
     AND actor.tombstone = false
    JOIN workspace
      ON actor.workspace_id = workspace.id
     AND workspace.tombstone = false
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    LEFT JOIN scoped_configuration
      ON scoped_configuration.scope_id = actor.id
     AND scoped_configuration.key = 'connector_version'
     AND scoped_configuration.scope_type = 'actor'
    WHERE
         jobs.config_type = 'sync'
     AND jobs.updated_at >= :cutoff_date
    ORDER BY
         jobs.updated_at DESC
    LIMIT :limit
    """
)

# Same as above but filtered to only successful syncs
SELECT_RECENT_SUCCESSFUL_SYNCS_FOR_SOURCE_CONNECTOR = sqlalchemy.text(
    """
    SELECT
         jobs.id AS job_id,
         jobs.scope AS connection_id,
         jobs.status AS job_status,
         jobs.started_at AS job_started_at,
         jobs.updated_at AS job_updated_at,
         connection.name AS connection_name,
         actor.id AS actor_id,
         actor.name AS actor_name,
         actor.actor_definition_id,
         actor.tombstone AS actor_tombstone,
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name,
         scoped_configuration.origin_type AS pin_origin_type,
         scoped_configuration.origin AS pin_origin,
         scoped_configuration.value AS pinned_version_id
    FROM jobs
    JOIN connection
      ON jobs.scope = connection.id::text
     AND connection.status != 'deprecated'
    JOIN actor
      ON connection.source_id = actor.id
     AND actor.actor_definition_id = :connector_definition_id
     AND actor.tombstone = false
    JOIN workspace
      ON actor.workspace_id = workspace.id
     AND workspace.tombstone = false
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    LEFT JOIN scoped_configuration
      ON scoped_configuration.scope_id = actor.id
     AND scoped_configuration.key = 'connector_version'
     AND scoped_configuration.scope_type = 'actor'
    WHERE
         jobs.config_type = 'sync'
     AND jobs.status = 'succeeded'
     AND jobs.updated_at >= :cutoff_date
    ORDER BY
         jobs.updated_at DESC
    LIMIT :limit
    """
)

# Same as above but filtered to only failed syncs
SELECT_RECENT_FAILED_SYNCS_FOR_SOURCE_CONNECTOR = sqlalchemy.text(
    """
    SELECT
         jobs.id AS job_id,
         jobs.scope AS connection_id,
         jobs.status AS job_status,
         jobs.started_at AS job_started_at,
         jobs.updated_at AS job_updated_at,
         connection.name AS connection_name,
         actor.id AS actor_id,
         actor.name AS actor_name,
         actor.actor_definition_id,
         actor.tombstone AS actor_tombstone,
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name,
         scoped_configuration.origin_type AS pin_origin_type,
         scoped_configuration.origin AS pin_origin,
         scoped_configuration.value AS pinned_version_id
    FROM jobs
    JOIN connection
      ON jobs.scope = connection.id::text
     AND connection.status != 'deprecated'
    JOIN actor
      ON connection.source_id = actor.id
     AND actor.actor_definition_id = :connector_definition_id
     AND actor.tombstone = false
    JOIN workspace
      ON actor.workspace_id = workspace.id
     AND workspace.tombstone = false
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    LEFT JOIN scoped_configuration
      ON scoped_configuration.scope_id = actor.id
     AND scoped_configuration.key = 'connector_version'
     AND scoped_configuration.scope_type = 'actor'
    WHERE
         jobs.config_type = 'sync'
     AND jobs.status = 'failed'
     AND jobs.updated_at >= :cutoff_date
    ORDER BY
         jobs.updated_at DESC
    LIMIT :limit
    """
)

# Get recent sync results for ALL actors using a DESTINATION connector definition.
SELECT_RECENT_SYNCS_FOR_DESTINATION_CONNECTOR = sqlalchemy.text(
    """
    SELECT
         jobs.id AS job_id,
         jobs.scope AS connection_id,
         jobs.status AS job_status,
         jobs.started_at AS job_started_at,
         jobs.updated_at AS job_updated_at,
         connection.name AS connection_name,
         actor.id AS actor_id,
         actor.name AS actor_name,
         actor.actor_definition_id,
         actor.tombstone AS actor_tombstone,
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name,
         scoped_configuration.origin_type AS pin_origin_type,
         scoped_configuration.origin AS pin_origin,
         scoped_configuration.value AS pinned_version_id
    FROM jobs
    JOIN connection
      ON jobs.scope = connection.id::text
     AND connection.status != 'deprecated'
    JOIN actor
      ON connection.destination_id = actor.id
     AND actor.actor_definition_id = :connector_definition_id
     AND actor.tombstone = false
    JOIN workspace
      ON actor.workspace_id = workspace.id
     AND workspace.tombstone = false
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    LEFT JOIN scoped_configuration
      ON scoped_configuration.scope_id = actor.id
     AND scoped_configuration.key = 'connector_version'
     AND scoped_configuration.scope_type = 'actor'
    WHERE
         jobs.config_type = 'sync'
     AND jobs.updated_at >= :cutoff_date
    ORDER BY
         jobs.updated_at DESC
    LIMIT :limit
    """
)

# Same as above but filtered to only successful syncs
SELECT_RECENT_SUCCESSFUL_SYNCS_FOR_DESTINATION_CONNECTOR = sqlalchemy.text(
    """
    SELECT
         jobs.id AS job_id,
         jobs.scope AS connection_id,
         jobs.status AS job_status,
         jobs.started_at AS job_started_at,
         jobs.updated_at AS job_updated_at,
         connection.name AS connection_name,
         actor.id AS actor_id,
         actor.name AS actor_name,
         actor.actor_definition_id,
         actor.tombstone AS actor_tombstone,
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name,
         scoped_configuration.origin_type AS pin_origin_type,
         scoped_configuration.origin AS pin_origin,
         scoped_configuration.value AS pinned_version_id
    FROM jobs
    JOIN connection
      ON jobs.scope = connection.id::text
     AND connection.status != 'deprecated'
    JOIN actor
      ON connection.destination_id = actor.id
     AND actor.actor_definition_id = :connector_definition_id
     AND actor.tombstone = false
    JOIN workspace
      ON actor.workspace_id = workspace.id
     AND workspace.tombstone = false
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    LEFT JOIN scoped_configuration
      ON scoped_configuration.scope_id = actor.id
     AND scoped_configuration.key = 'connector_version'
     AND scoped_configuration.scope_type = 'actor'
    WHERE
         jobs.config_type = 'sync'
     AND jobs.status = 'succeeded'
     AND jobs.updated_at >= :cutoff_date
    ORDER BY
         jobs.updated_at DESC
    LIMIT :limit
    """
)

# Same as above but filtered to only failed syncs
SELECT_RECENT_FAILED_SYNCS_FOR_DESTINATION_CONNECTOR = sqlalchemy.text(
    """
    SELECT
         jobs.id AS job_id,
         jobs.scope AS connection_id,
         jobs.status AS job_status,
         jobs.started_at AS job_started_at,
         jobs.updated_at AS job_updated_at,
         connection.name AS connection_name,
         actor.id AS actor_id,
         actor.name AS actor_name,
         actor.actor_definition_id,
         actor.tombstone AS actor_tombstone,
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name,
         scoped_configuration.origin_type AS pin_origin_type,
         scoped_configuration.origin AS pin_origin,
         scoped_configuration.value AS pinned_version_id
    FROM jobs
    JOIN connection
      ON jobs.scope = connection.id::text
     AND connection.status != 'deprecated'
    JOIN actor
      ON connection.destination_id = actor.id
     AND actor.actor_definition_id = :connector_definition_id
     AND actor.tombstone = false
    JOIN workspace
      ON actor.workspace_id = workspace.id
     AND workspace.tombstone = false
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    LEFT JOIN scoped_configuration
      ON scoped_configuration.scope_id = actor.id
     AND scoped_configuration.key = 'connector_version'
     AND scoped_configuration.scope_type = 'actor'
    WHERE
         jobs.config_type = 'sync'
     AND jobs.status = 'failed'
     AND jobs.updated_at >= :cutoff_date
    ORDER BY
         jobs.updated_at DESC
    LIMIT :limit
    """
)

# Get failed attempt results for ALL actors using a connector definition.
# Finds all actors with the given actor_definition_id and returns their failed sync attempts,
# regardless of whether they have explicit version pins.
# Query starts from attempts table to leverage indexed columns (ended_at, status).
# Note: This query only supports SOURCE connectors (joins via connection.source_id).
# The LEFT JOIN to scoped_configuration provides pin context when available (pin_origin_type,
# pin_origin, pinned_version_id will be NULL for unpinned actors).
SELECT_FAILED_SYNC_ATTEMPTS_FOR_CONNECTOR = sqlalchemy.text(
    """
    SELECT
         jobs.id AS job_id,
         jobs.scope AS connection_id,
         jobs.status AS latest_job_status,
         jobs.started_at AS job_started_at,
         jobs.updated_at AS job_updated_at,
         connection.name AS connection_name,
         actor.id AS actor_id,
         actor.name AS actor_name,
         actor.actor_definition_id,
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name,
         scoped_configuration.origin_type AS pin_origin_type,
         scoped_configuration.origin AS pin_origin,
         scoped_configuration.value AS pinned_version_id,
         attempts.id AS failed_attempt_id,
         attempts.attempt_number AS failed_attempt_number,
         attempts.status AS failed_attempt_status,
         attempts.created_at AS failed_attempt_created_at,
         attempts.ended_at AS failed_attempt_ended_at,
         attempts.failure_summary,
         attempts.processing_task_queue
    FROM attempts
    JOIN jobs
      ON jobs.id = attempts.job_id
     AND jobs.config_type = 'sync'
     AND jobs.updated_at >= :cutoff_date
    JOIN connection
      ON jobs.scope = connection.id::text
    JOIN actor
      ON connection.source_id = actor.id
     AND actor.actor_definition_id = :connector_definition_id
    JOIN workspace
      ON actor.workspace_id = workspace.id
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    LEFT JOIN scoped_configuration
      ON scoped_configuration.scope_id = actor.id
     AND scoped_configuration.key = 'connector_version'
     AND scoped_configuration.scope_type = 'actor'
    WHERE
         attempts.ended_at >= :cutoff_date
     AND attempts.status = 'failed'
    ORDER BY
         attempts.ended_at DESC
    LIMIT :limit
    """
)

# =============================================================================
# Dataplane and Workspace Queries
# =============================================================================

# List all dataplane groups with workspace counts
SELECT_DATAPLANES_LIST = sqlalchemy.text(
    """
    SELECT
         dataplane_group.id AS dataplane_group_id,
         dataplane_group.name AS dataplane_name,
         dataplane_group.organization_id,
         dataplane_group.enabled,
         dataplane_group.tombstone,
         dataplane_group.created_at,
         COUNT(workspace.id) AS workspace_count
    FROM dataplane_group
    LEFT JOIN workspace
      ON workspace.dataplane_group_id = dataplane_group.id
     AND workspace.tombstone = false
    WHERE
         dataplane_group.tombstone = false
    GROUP BY
         dataplane_group.id,
         dataplane_group.name,
         dataplane_group.organization_id,
         dataplane_group.enabled,
         dataplane_group.tombstone,
         dataplane_group.created_at
    ORDER BY
         workspace_count DESC
    """
)

# Get workspace info including dataplane group for EU filtering
SELECT_WORKSPACE_INFO = sqlalchemy.text(
    """
    SELECT
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.slug,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name,
         workspace.created_at,
         workspace.tombstone
    FROM workspace
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    WHERE
         workspace.id = :workspace_id
    """
)

# Get all workspaces in an organization with dataplane info
SELECT_ORG_WORKSPACES = sqlalchemy.text(
    """
    SELECT
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.slug,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name,
         workspace.created_at,
         workspace.tombstone
    FROM workspace
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    WHERE
         workspace.organization_id = :organization_id
     AND workspace.tombstone = false
    ORDER BY
         workspace.name
    """
)

# =============================================================================
# Workspace Lookup by Email Domain
# =============================================================================

# Find workspaces by email domain
# This is useful for identifying workspaces based on user email domains
# (e.g., finding partner accounts like MotherDuck by searching for "motherduck.com")
SELECT_WORKSPACES_BY_EMAIL_DOMAIN = sqlalchemy.text(
    """
    SELECT DISTINCT
         workspace.organization_id,
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.slug,
         workspace.email,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name,
         workspace.created_at
    FROM workspace
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    WHERE
         workspace.email LIKE '%@' || :email_domain
     AND workspace.tombstone = false
    ORDER BY
         workspace.organization_id,
         workspace.name
    LIMIT :limit
    """
)

# =============================================================================
# Connector Connection Stats Queries (Aggregate Counts)
# =============================================================================

# Count connections by SOURCE connector with latest attempt status breakdown
# Groups by pinned version and provides counts of succeeded/failed/other attempts
# Uses a CTE to get the latest attempt per connection, then aggregates
SELECT_SOURCE_CONNECTION_STATS = sqlalchemy.text(
    """
    WITH latest_attempts AS (
        SELECT DISTINCT ON (connection.id)
            connection.id AS connection_id,
            connection.status AS connection_status,
            scoped_configuration.value AS pinned_version_id,
            attempts.status::text AS latest_attempt_status
        FROM connection
        JOIN actor
          ON connection.source_id = actor.id
         AND actor.actor_definition_id = :connector_definition_id
         AND actor.tombstone = false
        JOIN workspace
          ON actor.workspace_id = workspace.id
         AND workspace.tombstone = false
        LEFT JOIN scoped_configuration
          ON scoped_configuration.scope_id = actor.id
         AND scoped_configuration.key = 'connector_version'
         AND scoped_configuration.scope_type = 'actor'
        LEFT JOIN jobs
          ON jobs.scope = connection.id::text
         AND jobs.config_type = 'sync'
         AND jobs.updated_at >= :cutoff_date
        LEFT JOIN attempts
          ON attempts.job_id = jobs.id
        WHERE
             connection.status != 'deprecated'
        ORDER BY
             connection.id,
             attempts.ended_at DESC NULLS LAST
    )
    SELECT
        pinned_version_id,
        COUNT(*) AS total_connections,
        COUNT(*) FILTER (WHERE connection_status = 'active') AS enabled_connections,
        COUNT(*) FILTER (WHERE latest_attempt_status IS NOT NULL) AS active_connections,
        COUNT(*) FILTER (WHERE pinned_version_id IS NOT NULL) AS pinned_connections,
        COUNT(*) FILTER (WHERE pinned_version_id IS NULL) AS unpinned_connections,
        COUNT(*) FILTER (WHERE latest_attempt_status = 'succeeded') AS succeeded_connections,
        COUNT(*) FILTER (WHERE latest_attempt_status = 'failed') AS failed_connections,
        COUNT(*) FILTER (WHERE latest_attempt_status = 'cancelled') AS cancelled_connections,
        COUNT(*) FILTER (WHERE latest_attempt_status = 'running') AS running_connections,
        COUNT(*) FILTER (WHERE latest_attempt_status IS NULL) AS unknown_connections
    FROM latest_attempts
    GROUP BY pinned_version_id
    ORDER BY total_connections DESC
    """
)

# Count connections by DESTINATION connector with latest attempt status breakdown
SELECT_DESTINATION_CONNECTION_STATS = sqlalchemy.text(
    """
    WITH latest_attempts AS (
        SELECT DISTINCT ON (connection.id)
            connection.id AS connection_id,
            connection.status AS connection_status,
            scoped_configuration.value AS pinned_version_id,
            attempts.status::text AS latest_attempt_status
        FROM connection
        JOIN actor
          ON connection.destination_id = actor.id
         AND actor.actor_definition_id = :connector_definition_id
         AND actor.tombstone = false
        JOIN workspace
          ON actor.workspace_id = workspace.id
         AND workspace.tombstone = false
        LEFT JOIN scoped_configuration
          ON scoped_configuration.scope_id = actor.id
         AND scoped_configuration.key = 'connector_version'
         AND scoped_configuration.scope_type = 'actor'
        LEFT JOIN jobs
          ON jobs.scope = connection.id::text
         AND jobs.config_type = 'sync'
         AND jobs.updated_at >= :cutoff_date
        LEFT JOIN attempts
          ON attempts.job_id = jobs.id
        WHERE
             connection.status != 'deprecated'
        ORDER BY
             connection.id,
             attempts.ended_at DESC NULLS LAST
    )
    SELECT
        pinned_version_id,
        COUNT(*) AS total_connections,
        COUNT(*) FILTER (WHERE connection_status = 'active') AS enabled_connections,
        COUNT(*) FILTER (WHERE latest_attempt_status IS NOT NULL) AS active_connections,
        COUNT(*) FILTER (WHERE pinned_version_id IS NOT NULL) AS pinned_connections,
        COUNT(*) FILTER (WHERE pinned_version_id IS NULL) AS unpinned_connections,
        COUNT(*) FILTER (WHERE latest_attempt_status = 'succeeded') AS succeeded_connections,
        COUNT(*) FILTER (WHERE latest_attempt_status = 'failed') AS failed_connections,
        COUNT(*) FILTER (WHERE latest_attempt_status = 'cancelled') AS cancelled_connections,
        COUNT(*) FILTER (WHERE latest_attempt_status = 'running') AS running_connections,
        COUNT(*) FILTER (WHERE latest_attempt_status IS NULL) AS unknown_connections
    FROM latest_attempts
    GROUP BY pinned_version_id
    ORDER BY total_connections DESC
    """
)

# =============================================================================
# Stream-based Connection Queries
# =============================================================================

# Query connections by source connector type that have a specific stream enabled
# The catalog field is JSONB with structure: {"streams": [{"stream": {"name": "..."}, ...}, ...]}
SELECT_CONNECTIONS_BY_SOURCE_CONNECTOR_AND_STREAM = sqlalchemy.text(
    """
    SELECT
         connection.id AS connection_id,
         connection.name AS connection_name,
         connection.source_id,
         connection.status AS connection_status,
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name,
         source_actor.actor_definition_id AS source_definition_id,
         source_actor.name AS source_name
    FROM connection
    JOIN actor AS source_actor
      ON connection.source_id = source_actor.id
     AND source_actor.tombstone = false
    JOIN workspace
      ON source_actor.workspace_id = workspace.id
     AND workspace.tombstone = false
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    WHERE
         source_actor.actor_definition_id = :connector_definition_id
     AND connection.status = 'active'
     AND EXISTS (
         SELECT 1 FROM jsonb_array_elements(connection.catalog->'streams') AS stream
         WHERE stream->'stream'->>'name' = :stream_name
     )
    LIMIT :limit
    """
)

# Query connections by source connector type and stream, filtered by organization
SELECT_CONNECTIONS_BY_SOURCE_CONNECTOR_AND_STREAM_AND_ORG = sqlalchemy.text(
    """
    SELECT
         connection.id AS connection_id,
         connection.name AS connection_name,
         connection.source_id,
         connection.status AS connection_status,
         workspace.id AS workspace_id,
         workspace.name AS workspace_name,
         workspace.organization_id,
         workspace.dataplane_group_id,
         dataplane_group.name AS dataplane_name,
         source_actor.actor_definition_id AS source_definition_id,
         source_actor.name AS source_name
    FROM connection
    JOIN actor AS source_actor
      ON connection.source_id = source_actor.id
     AND source_actor.tombstone = false
    JOIN workspace
      ON source_actor.workspace_id = workspace.id
     AND workspace.tombstone = false
    LEFT JOIN dataplane_group
      ON workspace.dataplane_group_id = dataplane_group.id
    WHERE
         source_actor.actor_definition_id = :connector_definition_id
     AND workspace.organization_id = :organization_id
     AND connection.status = 'active'
     AND EXISTS (
         SELECT 1 FROM jsonb_array_elements(connection.catalog->'streams') AS stream
         WHERE stream->'stream'->>'name' = :stream_name
     )
    LIMIT :limit
    """
)

# =============================================================================
# Connector Rollout Queries
# =============================================================================

# List all rollouts for a connector definition, with version details
SELECT_CONNECTOR_ROLLOUTS = sqlalchemy.text(
    """
    SELECT
         cr.id AS rollout_id,
         cr.actor_definition_id,
         cr.state,
         cr.initial_rollout_pct,
         cr.current_target_rollout_pct,
         cr.final_target_rollout_pct,
         cr.has_breaking_changes,
         cr.max_step_wait_time_mins,
         cr.rollout_strategy,
         cr.workflow_run_id,
         cr.error_msg,
         cr.failed_reason,
         cr.paused_reason,
         cr.tag,
         cr.created_at,
         cr.updated_at,
         cr.completed_at,
         cr.expires_at,
         rc_version.docker_image_tag AS rc_docker_image_tag,
         rc_version.docker_repository AS rc_docker_repository,
         initial_version.docker_image_tag AS initial_docker_image_tag,
         initial_version.docker_repository AS initial_docker_repository
    FROM connector_rollout cr
    JOIN actor_definition_version rc_version
      ON cr.release_candidate_version_id = rc_version.id
    LEFT JOIN actor_definition_version initial_version
      ON cr.initial_version_id = initial_version.id
    WHERE
         cr.actor_definition_id = :actor_definition_id
    ORDER BY
         cr.created_at DESC
    LIMIT :limit
    """
)

# List all active (non-terminal) rollouts across all connectors
# Active states: initialized, workflow_started, in_progress, paused, finalizing, errored
SELECT_ACTIVE_CONNECTOR_ROLLOUTS = sqlalchemy.text(
    """
    SELECT
         cr.id AS rollout_id,
         cr.actor_definition_id,
         cr.state,
         cr.initial_rollout_pct,
         cr.current_target_rollout_pct,
         cr.final_target_rollout_pct,
         cr.has_breaking_changes,
         cr.max_step_wait_time_mins,
         cr.rollout_strategy,
         cr.workflow_run_id,
         cr.error_msg,
         cr.failed_reason,
         cr.paused_reason,
         cr.tag,
         cr.created_at,
         cr.updated_at,
         cr.completed_at,
         cr.expires_at,
         rc_version.docker_image_tag AS rc_docker_image_tag,
         rc_version.docker_repository AS rc_docker_repository,
         initial_version.docker_image_tag AS initial_docker_image_tag,
         initial_version.docker_repository AS initial_docker_repository
    FROM connector_rollout cr
    JOIN actor_definition_version rc_version
      ON cr.release_candidate_version_id = rc_version.id
    LEFT JOIN actor_definition_version initial_version
      ON cr.initial_version_id = initial_version.id
    WHERE
         cr.state IN ('initialized', 'workflow_started', 'in_progress', 'paused', 'finalizing', 'errored')
    ORDER BY
         cr.created_at DESC
    LIMIT :limit
    """
)

# List active (non-terminal) rollouts for a specific connector definition
# Active states: initialized, workflow_started, in_progress, paused, finalizing, errored
SELECT_ACTIVE_CONNECTOR_ROLLOUTS_BY_DEFINITION = sqlalchemy.text(
    """
    SELECT
         cr.id AS rollout_id,
         cr.actor_definition_id,
         cr.state,
         cr.initial_rollout_pct,
         cr.current_target_rollout_pct,
         cr.final_target_rollout_pct,
         cr.has_breaking_changes,
         cr.max_step_wait_time_mins,
         cr.rollout_strategy,
         cr.workflow_run_id,
         cr.error_msg,
         cr.failed_reason,
         cr.paused_reason,
         cr.tag,
         cr.created_at,
         cr.updated_at,
         cr.completed_at,
         cr.expires_at,
         rc_version.docker_image_tag AS rc_docker_image_tag,
         rc_version.docker_repository AS rc_docker_repository,
         initial_version.docker_image_tag AS initial_docker_image_tag,
         initial_version.docker_repository AS initial_docker_repository
    FROM connector_rollout cr
    JOIN actor_definition_version rc_version
      ON cr.release_candidate_version_id = rc_version.id
    LEFT JOIN actor_definition_version initial_version
      ON cr.initial_version_id = initial_version.id
    WHERE
         cr.actor_definition_id = :actor_definition_id
     AND cr.state IN ('initialized', 'workflow_started', 'in_progress', 'paused', 'finalizing', 'errored')
    ORDER BY
         cr.created_at DESC
    LIMIT :limit
    """
)

# Get a specific rollout by ID
SELECT_CONNECTOR_ROLLOUT_BY_ID = sqlalchemy.text(
    """
    SELECT
         cr.id AS rollout_id,
         cr.actor_definition_id,
         cr.state,
         cr.initial_rollout_pct,
         cr.current_target_rollout_pct,
         cr.final_target_rollout_pct,
         cr.has_breaking_changes,
         cr.max_step_wait_time_mins,
         cr.rollout_strategy,
         cr.workflow_run_id,
         cr.error_msg,
         cr.failed_reason,
         cr.paused_reason,
         cr.filters,
         cr.tag,
         cr.created_at,
         cr.updated_at,
         cr.completed_at,
         cr.expires_at,
         rc_version.docker_image_tag AS rc_docker_image_tag,
         rc_version.docker_repository AS rc_docker_repository,
         initial_version.docker_image_tag AS initial_docker_image_tag,
         initial_version.docker_repository AS initial_docker_repository
    FROM connector_rollout cr
    JOIN actor_definition_version rc_version
      ON cr.release_candidate_version_id = rc_version.id
    LEFT JOIN actor_definition_version initial_version
      ON cr.initial_version_id = initial_version.id
    WHERE
         cr.id = :rollout_id
    """
)

# =============================================================================
# Connector Rollout Monitoring Queries
# =============================================================================

# Get actors pinned to a specific rollout (via scoped_configuration.origin = rollout_id)
# This shows which actors are currently participating in the rollout
SELECT_ACTORS_PINNED_TO_ROLLOUT = sqlalchemy.text(
    """
    SELECT
         actor.id AS actor_id,
         actor.name AS actor_name,
         actor.actor_definition_id,
         actor.workspace_id,
         workspace.name AS workspace_name,
         workspace.organization_id,
         scoped_configuration.origin_type AS pin_origin_type,
         scoped_configuration.origin AS pin_origin,
         scoped_configuration.created_at AS pin_created_at,
         scoped_configuration.value AS pinned_version_id
    FROM scoped_configuration
    JOIN actor
      ON scoped_configuration.scope_id = actor.id
    JOIN workspace
      ON actor.workspace_id = workspace.id
    WHERE
         scoped_configuration.key = 'connector_version'
     AND scoped_configuration.scope_type = 'actor'
     AND scoped_configuration.origin = :rollout_id
    ORDER BY
         scoped_configuration.created_at DESC
    """
)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import functools
import json
import logging
import os
import re
from dataclasses import dataclass
from enum import Enum
from glob import glob
from pathlib import Path
from typing import List, Optional, Set, Tuple, Union

import git
import requests
import yaml
from pydash.collections import find
from pydash.objects import get
from rich.console import Console
from simpleeval import simple_eval

console = Console()

DIFFED_BRANCH = os.environ.get("DIFFED_BRANCH", "origin/master")
OSS_CATALOG_URL = "https://connectors.airbyte.com/files/registries/v0/oss_registry.json"
CLOUD_CATALOG_URL = (
    "https://connectors.airbyte.com/files/registries/v0/cloud_registry.json"
)
BASE_AIRBYTE_DOCS_URL = "https://docs.airbyte.com"
CONNECTOR_PATH_PREFIX = "airbyte-integrations/connectors"
SOURCE_CONNECTOR_PATH_PREFIX = CONNECTOR_PATH_PREFIX + "/source-"
DESTINATION_CONNECTOR_PATH_PREFIX = CONNECTOR_PATH_PREFIX + "/destination-"

THIRD_PARTY_GLOB = "third-party"
THIRD_PARTY_CONNECTOR_PATH_PREFIX = CONNECTOR_PATH_PREFIX + f"/{THIRD_PARTY_GLOB}/"
SCAFFOLD_CONNECTOR_GLOB = "-scaffold-"


ACCEPTANCE_TEST_CONFIG_FILE_NAME = "acceptance-test-config.yml"
METADATA_FILE_NAME = "metadata.yaml"
AIRBYTE_DOCKER_REPO = "airbyte"
AIRBYTE_REPO_DIRECTORY_NAME = "airbyte"
GRADLE_PROJECT_RE_PATTERN = r"project\((['\"])(.+?)\1\)"
TEST_GRADLE_DEPENDENCIES = [
    "testImplementation",
    "testCompileOnly",
    "integrationTestJavaImplementation",
    "performanceTestJavaImplementation",
    "testFixturesCompileOnly",
    "testFixturesImplementation",
]


def download_catalog(catalog_url):
    response = requests.get(catalog_url)
    response.raise_for_status()
    return response.json()


OSS_CATALOG = download_catalog(OSS_CATALOG_URL)
MANIFEST_FILE_NAME = "manifest.yaml"
COMPONENTS_FILE_NAME = "components.py"
DOCKERFILE_FILE_NAME = "Dockerfile"
PYPROJECT_FILE_NAME = "pyproject.toml"
ICON_FILE_NAME = "icon.svg"
POETRY_LOCK_FILE_NAME = "poetry.lock"

STRATEGIC_CONNECTOR_THRESHOLDS = {
    "sl": 200,
    "ql": 400,
}

ALLOWED_HOST_THRESHOLD = {
    "ql": 300,
}


class ConnectorInvalidNameError(Exception):
    pass


class ConnectorVersionNotFound(Exception):
    pass


def get_connector_name_from_path(path):
    return path.split("/")[2]


def get_changed_metadata(diff_regex: Optional[str] = None) -> Set[str]:
    """Retrieve the set of connectors for which the metadata file was changed in the current branch (compared to master).

    Args:
        diff_regex (str): Find the edited files that contain the following regex in their change.

    Returns:
        Set[Connector]: Set of connectors that were changed
    """
    return get_changed_file(METADATA_FILE_NAME, diff_regex)


def get_changed_file(file_name: str, diff_regex: Optional[str] = None) -> Set[str]:
    """Retrieve the set of connectors for which the given file was changed in the current branch (compared to master).

    Args:
        diff_regex (str): Find the edited files that contain the following regex in their change.

    Returns:
        Set[Connector]: Set of connectors that were changed
    """
    airbyte_repo = git.Repo(search_parent_directories=True)

    if diff_regex is None:
        diff_command_args = ("--name-only", DIFFED_BRANCH)
    else:
        diff_command_args = ("--name-only", f"-G{diff_regex}", DIFFED_BRANCH)

    changed_acceptance_test_config_paths = {
        file_path
        for file_path in airbyte_repo.git.diff(*diff_command_args).split("\n")
        if file_path.startswith(SOURCE_CONNECTOR_PATH_PREFIX)
        and file_path.endswith(file_name)
    }
    return {
        Connector(get_connector_name_from_path(changed_file))
        for changed_file in changed_acceptance_test_config_paths
    }


def has_local_cdk_ref(build_file: Path) -> bool:
    """Return true if the build file uses the local CDK.

    Args:
        build_file (Path): Path to the build.gradle/build.gradle.kts file of the project.

    Returns:
        bool: True if using local CDK.
    """
    contents = "\n".join(
        [
            # Return contents without inline code comments
            line.split("//")[0]
            for line in build_file.read_text().split("\n")
        ]
    )
    contents = contents.replace(" ", "")
    return "useLocalCdk=true" in contents


def get_gradle_dependencies_block(build_file: Path) -> str:
    """Get the dependencies block of a Gradle file.

    Args:
        build_file (Path): Path to the build.gradle/build.gradle.kts file of the project.

    Returns:
        str: The dependencies block of the Gradle file.
    """
    contents = build_file.read_text().split("\n")
    dependency_block = []
    in_dependencies_block = False
    for line in contents:
        if line.strip().startswith("dependencies"):
            in_dependencies_block = True
            continue
        if in_dependencies_block:
            if line.startswith("}"):
                in_dependencies_block = False
                break
            else:
                dependency_block.append(line)
    dependencies_block = "\n".join(dependency_block)
    return dependencies_block


def parse_gradle_dependencies(build_file: Path) -> Tuple[List[Path], List[Path]]:
    """Parse the dependencies block of a Gradle file and return the list of project dependencies and test dependencies.

    Args:
        build_file (Path): _description_

    Returns:
        Tuple[List[Tuple[str, Path]], List[Tuple[str, Path]]]: _description_
    """

    dependencies_block = get_gradle_dependencies_block(build_file)

    project_dependencies: List[Path] = []
    test_dependencies: List[Path] = []

    # Find all matches for test dependencies and regular dependencies
    matches = re.findall(
        r"(compileOnly|testCompileOnly|testFixturesCompileOnly|testFixturesImplementation|testImplementation|integrationTestJavaImplementation|performanceTestJavaImplementation|implementation|api).*?project\(['\"](.*?)['\"]\)",
        dependencies_block,
    )
    if matches:
        # Iterate through each match
        for match in matches:
            dependency_type, project_path = match
            path_parts = project_path.split(":")
            path = Path(*path_parts)

            if dependency_type in TEST_GRADLE_DEPENDENCIES:
                test_dependencies.append(path)
            else:
                project_dependencies.append(path)

    # Dedupe dependencies:
    project_dependencies = list(set(project_dependencies))
    test_dependencies = list(set(test_dependencies))

    return project_dependencies, test_dependencies


def get_all_gradle_dependencies(
    build_file: Path,
    with_test_dependencies: bool = True,
    found_dependencies: Optional[List[Path]] = None,
) -> List[Path]:
    """Recursively retrieve all transitive dependencies of a Gradle project.

    Args:
        build_file (Path): Path to the build.gradle/build.gradle.kts file of the project.
        found_dependencies (List[Path]): List of dependencies that have already been found. Defaults to None.

    Returns:
        List[Path]: All dependencies of the project.
    """
    if found_dependencies is None:
        found_dependencies = []
    project_dependencies, test_dependencies = parse_gradle_dependencies(build_file)
    all_dependencies = (
        project_dependencies + test_dependencies
        if with_test_dependencies
        else project_dependencies
    )
    valid_build_files = ["build.gradle", "build.gradle.kts"]
    for dependency_path in all_dependencies:
        for build_file in valid_build_files:
            if (
                dependency_path not in found_dependencies
                and Path(dependency_path / build_file).exists()
            ):
                found_dependencies.append(dependency_path)
                get_all_gradle_dependencies(
                    dependency_path / build_file,
                    with_test_dependencies,
                    found_dependencies,
                )

    return found_dependencies


class ConnectorLanguage(str, Enum):
    PYTHON = "python"
    JAVA = "java"
    LOW_CODE = "low-code"
    MANIFEST_ONLY = "manifest-only"


class ConnectorLanguageError(Exception):
    pass


@dataclass(frozen=True)
class Connector:
    """Utility class to gather metadata about a connector."""

    # Path to the connector directory relative to the CONNECTOR_PATH_PREFIX
    # e.g source-google-sheets or third-party/farosai/airbyte-pagerduty-source
    relative_connector_path: str

    def _get_type_and_name_from_technical_name(self) -> Tuple[str, str]:
        if "-" not in self.technical_name:
            raise ConnectorInvalidNameError(
                f"Connector type and name could not be inferred from {self.technical_name}"
            )
        _type = self.technical_name.split("-")[0]
        name = self.technical_name[len(_type) + 1 :]
        return _type, name

    @property
    def technical_name(self) -> str:
        """
        Return the technical name of the connector from the given relative_connector_path
        e.g. source-google-sheets -> source-google-sheets or third-party/farosai/airbyte-pagerduty-source -> airbyte-pagerduty-source
        """
        return self.relative_connector_path.split("/")[-1]

    @property
    def name(self):
        return self._get_type_and_name_from_technical_name()[1]

    @property
    def connector_type(self) -> str:
        return self.metadata["connectorType"] if self.metadata else None

    @property
    def is_third_party(self) -> bool:
        return THIRD_PARTY_GLOB in self.relative_connector_path

    @property
    def has_airbyte_docs(self) -> bool:
        return (
            self.metadata
            and self.metadata.get("documentationUrl") is not None
            and BASE_AIRBYTE_DOCS_URL in str(self.metadata.get("documentationUrl"))
        )

    @property
    def local_connector_documentation_directory(self) -> Path:
        return Path(f"./docs/integrations/{self.connector_type}s")

    @property
    def relative_documentation_path_str(self) -> str:
        documentation_url = self.metadata["documentationUrl"]
        relative_documentation_path = documentation_url.replace(
            BASE_AIRBYTE_DOCS_URL, ""
        )

        # strip leading and trailing slashes
        relative_documentation_path = relative_documentation_path.strip("/")

        return f"./docs/{relative_documentation_path}"

    @property
    def documentation_file_name(self) -> str:
        return self.metadata.get("documentationUrl").split("/")[-1] + ".md"

    @property
    def documentation_file_path(self) -> Optional[Path]:
        return (
            Path(f"{self.relative_documentation_path_str}.md")
            if self.has_airbyte_docs
            else None
        )

    @property
    def inapp_documentation_file_path(self) -> Path:
        if not self.has_airbyte_docs:
            return None

        return Path(f"{self.relative_documentation_path_str}.inapp.md")

    @property
    def migration_guide_file_name(self) -> str:
        return f"{self.name}-migrations.md"

    @property
    def migration_guide_file_path(self) -> Path:
        return (
            self.local_connector_documentation_directory
            / self.migration_guide_file_name
        )

    @property
    def icon_path(self) -> Path:
        file_path = self.code_directory / ICON_FILE_NAME
        return file_path

    @property
    def code_directory(self) -> Path:
        return Path(f"./{CONNECTOR_PATH_PREFIX}/{self.relative_connector_path}")

    @property
    def python_source_dir_path(self) -> Path:
        return self.code_directory / self.technical_name.replace("-", "_")

    @property
    def _manifest_only_path(self) -> Path:
        return self.code_directory / MANIFEST_FILE_NAME

    @property
    def _manifest_low_code_path(self) -> Path:
        return self.python_source_dir_path / MANIFEST_FILE_NAME

    @property
    def manifest_path(self) -> Path:
        if self._manifest_only_path.is_file():
            return self._manifest_only_path

        return self._manifest_low_code_path

    @property
    def manifest_only_components_path(self) -> Path:
        """Return the path to the components.py file of a manifest-only connector."""
        return self.code_directory / COMPONENTS_FILE_NAME

    @property
    def has_dockerfile(self) -> bool:
        return self.dockerfile_file_path.is_file()

    @property
    def dockerfile_file_path(self) -> Path:
        return self.code_directory / DOCKERFILE_FILE_NAME

    @property
    def pyproject_file_path(self) -> Path:
        return self.code_directory / PYPROJECT_FILE_NAME

    @property
    def metadata_file_path(self) -> Path:
        return self.code_directory / METADATA_FILE_NAME

    @property
    def metadata(self) -> Optional[dict]:
        file_path = self.metadata_file_path
        if not file_path.is_file():
            return None
        return yaml.safe_load((self.code_directory / METADATA_FILE_NAME).read_text())[
            "data"
        ]

    @property
    def connector_spec_file_content(self) -> Optional[dict]:
        """
        The spec source of truth is the actual output of the spec command, as connector can mutate their spec.
        But this is the best effort approach at statically fetching a spec without running the command on the connector.
        Which is "good enough" in some cases.
        """
        yaml_spec = Path(self.python_source_dir_path / "spec.yaml")
        json_spec = Path(self.python_source_dir_path / "spec.json")

        if yaml_spec.exists():
            return yaml.safe_load(yaml_spec.read_text())
        elif json_spec.exists():
            with open(json_spec) as f:
                return json.load(f)
        elif self.manifest_path.exists():
            return yaml.safe_load(self.manifest_path.read_text())["spec"]

        return None

    @property
    def language(self) -> ConnectorLanguage:
        if Path(self.code_directory / "manifest.yaml").is_file():
            return ConnectorLanguage.MANIFEST_ONLY
        if Path(
            self.code_directory
            / self.technical_name.replace("-", "_")
            / "manifest.yaml"
        ).is_file():
            return ConnectorLanguage.LOW_CODE
        if (
            Path(self.code_directory / "setup.py").is_file()
            or Path(self.code_directory / "pyproject.toml").is_file()
        ):
            return ConnectorLanguage.PYTHON
        if (
            Path(self.code_directory / "src" / "main" / "java").exists()
            or Path(self.code_directory / "src" / "main" / "kotlin").exists()
        ):
            return ConnectorLanguage.JAVA
        return None

    @property
    def version(self) -> Optional[str]:
        if self.metadata is None:
            return self.version_in_dockerfile_label
        return self.metadata["dockerImageTag"]

    @property
    def version_in_dockerfile_label(self) -> Optional[str]:
        if not self.has_dockerfile:
            return None
        with open(self.code_directory / "Dockerfile") as f:
            for line in f:
                if "io.airbyte.version" in line:
                    return line.split("=")[1].strip()
        raise ConnectorVersionNotFound(
            """
            Could not find the connector version from its Dockerfile.
            The io.airbyte.version tag is missing.
            """
        )

    @property
    def name_from_metadata(self) -> Optional[str]:
        return self.metadata.get("name") if self.metadata else None

    @property
    def support_level(self) -> Optional[str]:
        return self.metadata.get("supportLevel") if self.metadata else None

    def metadata_query_match(self, query_string: str) -> bool:
        """Evaluate a query string against the connector metadata.

        Based on the simpleeval library:
        https://github.com/danthedeckie/simpleeval

        Examples
        --------
        >>> connector.metadata_query_match("'s3' in data.name")
        True

        >>> connector.metadata_query_match("data.supportLevel == 'certified'")
        False

        >>> connector.metadata_query_match("data.ab_internal.ql >= 100")
        True

        Args:
            query_string (str): The query string to evaluate.

        Returns:
            bool: True if the query string matches the connector metadata, False otherwise.
        """
        try:
            matches = simple_eval(query_string, names={"data": self.metadata})
            return bool(matches)
        except Exception as e:
            # Skip on error as we not all fields are present in all connectors.
            logging.debug(
                f"Failed to evaluate query string {query_string} for connector {self.technical_name}, error: {e}"
            )
            return False

    @property
    def ab_internal_sl(self) -> int:
        """Airbyte Internal Field.

        More info can be found here: https://www.notion.so/Internal-Metadata-Fields-32b02037e7b244b7934214019d0b7cc9

        Returns:
            int: The value
        """
        default_value = 100
        sl_value = get(self.metadata, "ab_internal.sl")

        if sl_value is None:
            logging.warning(
                f"Connector {self.technical_name} does not have a `ab_internal.sl` defined in metadata.yaml. Defaulting to {default_value}"
            )
            return default_value

        return sl_value

    @property
    def ab_internal_ql(self) -> int:
        """Airbyte Internal Field.

        More info can be found here: https://www.notion.so/Internal-Metadata-Fields-32b02037e7b244b7934214019d0b7cc9

        Returns:
            int: The value
        """
        default_value = 100
        ql_value = get(self.metadata, "ab_internal.ql")

        if ql_value is None:
            logging.warning(
                f"Connector {self.technical_name} does not have a `ab_internal.ql` defined in metadata.yaml. Defaulting to {default_value}"
            )
            return default_value

        return ql_value

    @property
    def is_strategic_connector(self) -> bool:
        """Check if a connector qualifies as a strategic connector.

        Returns:
            bool: True if the connector is a high value connector, False otherwise.
        """
        if self.ab_internal_sl >= STRATEGIC_CONNECTOR_THRESHOLDS["sl"]:
            return True

        if self.ab_internal_ql >= STRATEGIC_CONNECTOR_THRESHOLDS["ql"]:
            return True

        return False

    @property
    def requires_high_test_strictness_level(self) -> bool:
        """Check if a connector requires high strictness CAT tests.

        Returns:
            bool: True if the connector requires high test strictness level, False otherwise.
        """
        return self.ab_internal_ql >= STRATEGIC_CONNECTOR_THRESHOLDS["ql"]

    @property
    def requires_allowed_hosts_check(self) -> bool:
        """Check if a connector requires allowed hosts.

        Returns:
            bool: True if the connector requires allowed hosts, False otherwise.
        """
        return self.ab_internal_ql >= ALLOWED_HOST_THRESHOLD["ql"]

    @property
    def allowed_hosts(self) -> Optional[List[str]]:
        return self.metadata.get("allowedHosts") if self.metadata else None

    @property
    def suggested_streams(self) -> Optional[List[str]]:
        return self.metadata.get("suggestedStreams") if self.metadata else None

    @property
    def acceptance_test_config_path(self) -> Path:
        return self.code_directory / ACCEPTANCE_TEST_CONFIG_FILE_NAME

    @property
    def acceptance_test_config(self) -> Optional[dict]:
        try:
            with open(self.acceptance_test_config_path) as acceptance_test_config_file:
                return yaml.safe_load(acceptance_test_config_file)
        except FileNotFoundError:
            logging.warning(
                f"No {ACCEPTANCE_TEST_CONFIG_FILE_NAME} file found for {self.technical_name}"
            )
            return None

    @property
    def supports_normalization(self) -> bool:
        return self.metadata and self.metadata.get("normalizationConfig") is not None

    @property
    def normalization_repository(self) -> Optional[str]:
        if self.supports_normalization:
            return f"{self.metadata['normalizationConfig']['normalizationRepository']}"

    @property
    def normalization_tag(self) -> Optional[str]:
        if self.supports_normalization:
            return f"{self.metadata['normalizationConfig']['normalizationTag']}"

    @property
    def is_using_poetry(self) -> bool:
        return Path(self.code_directory / "pyproject.toml").exists()

    @property
    def registry_primary_key_field(self) -> str:
        """
        The primary key field of the connector in the registry.

        example:
        - source -> sourceDefinitionId
        - destination -> destinationDefinitionId
        """
        return f"{self.connector_type}DefinitionId"

    @property
    def is_enabled_in_any_registry(self) -> bool:
        """Check if the connector is enabled in the registry.

        Example:
          - {registries: null} -> false
          - {registries: {oss: {enabled: false }}} -> false
          - {registries: {oss: {enabled: true }}} -> true
          - {registries: {cloud: {enabled: true }}} -> true

        Returns:
            bool: True if the connector is enabled, False otherwise.
        """
        registries = self.metadata.get("registryOverrides")
        if not registries:
            return False

        for registry in registries.values():
            if registry.get("enabled"):
                return True

        return False

    @property
    def is_released(self) -> bool:
        """Pull the the OSS registry and check if it the current definition ID and docker image tag are in the registry.
        If there is a match it means the connector is released.
        We use the OSS registry as the source of truth for released connectors as the cloud registry can be a subset of the OSS registry.

        Returns:
            bool: True if the connector is released, False otherwise.
        """
        metadata = self.metadata
        registry = download_catalog(OSS_CATALOG_URL)
        for connector in registry[f"{self.connector_type}s"]:
            if (
                connector[self.registry_primary_key_field] == metadata["definitionId"]
                and connector["dockerImageTag"] == metadata["dockerImageTag"]
            ):
                return True
        return False

    @property
    def cloud_usage(self) -> Optional[str]:
        """Pull the cloud registry, check if the connector is in the registry and return the usage metrics.

        Returns:
            Optional[str]: The usage metrics of the connector, could be one of ["low", "medium", "high"] or None if the connector is not in the registry.
        """
        metadata = self.metadata
        definition_id = metadata.get("definitionId")
        cloud_registry = download_catalog(CLOUD_CATALOG_URL)

        all_connectors_of_type = cloud_registry[f"{self.connector_type}s"]
        connector_entry = find(
            all_connectors_of_type, {self.registry_primary_key_field: definition_id}
        )
        if not connector_entry:
            return None

        return get(connector_entry, "generated.metrics.cloud.usage")

    @property
    def sbom_url(self) -> Optional[str]:
        """
        Fetches SBOM URL from the connector definition in the OSS registry, if it exists, None otherwise.
        """
        metadata = self.metadata
        definition_id = metadata.get("definitionId")
        # We use the OSS registry as the source of truth for released connectors as the cloud registry can be a subset of the OSS registry.
        oss_registry = download_catalog(OSS_CATALOG_URL)

        all_connectors_of_type = oss_registry[f"{self.connector_type}s"]
        connector_entry = find(
            all_connectors_of_type, {self.registry_primary_key_field: definition_id}
        )
        if not connector_entry:
            return None

        return get(connector_entry, "generated.sbomUrl")

    @property
    def image_address(self) -> str:
        return f"{self.metadata['dockerRepository']}:{self.metadata['dockerImageTag']}"

    @property
    def cdk_name(self) -> str | None:
        try:
            return [
                tag.split(":")[-1]
                for tag in self.metadata["tags"]
                if tag.startswith("cdk:")
            ][0]
        except IndexError:
            return None

    @property
    def base_image_address(self) -> str | None:
        return self.metadata.get("connectorBuildOptions", {}).get("baseImage")

    @property
    def uses_base_image(self) -> bool:
        return self.base_image_address is not None

    @property
    def base_image_version(self) -> str | None:
        if not self.uses_base_image:
            return None
        return self.base_image_address.split(":")[1].split("@")[0]

    def __repr__(self) -> str:
        return self.technical_name

    @functools.lru_cache(maxsize=2)
    def get_local_dependency_paths(
        self, with_test_dependencies: bool = True
    ) -> Set[Path]:
        dependencies_paths = []
        build_script = "build.gradle"
        if Path(self.code_directory / "build.gradle.kts").exists():
            build_script = "build.gradle.kts"

        if self.language == ConnectorLanguage.JAVA:
            dependencies_paths += [
                Path("./airbyte-cdk/java/airbyte-cdk"),
                Path("./airbyte-cdk/bulk"),
            ]
            dependencies_paths += get_all_gradle_dependencies(
                self.code_directory / build_script,
                with_test_dependencies=with_test_dependencies,
            )
        return sorted(list(set(dependencies_paths)))


def get_changed_connectors(
    modified_files: Optional[Set[Union[str, Path]]] = None,
    source: bool = True,
    destination: bool = True,
    third_party: bool = True,
) -> Set[Connector]:
    """Retrieve a set of Connectors that were changed in the current branch (compared to master)."""
    if modified_files is None:
        airbyte_repo = git.Repo(search_parent_directories=True)
        modified_files = airbyte_repo.git.diff("--name-only", DIFFED_BRANCH).split("\n")

    prefix_to_check = []
    if source:
        prefix_to_check.append(SOURCE_CONNECTOR_PATH_PREFIX)
    if destination:
        prefix_to_check.append(DESTINATION_CONNECTOR_PATH_PREFIX)
    if third_party:
        prefix_to_check.append(THIRD_PARTY_CONNECTOR_PATH_PREFIX)

    changed_source_connector_files = {
        file_path
        for file_path in modified_files
        if any(file_path.startswith(prefix) for prefix in prefix_to_check)
        and SCAFFOLD_CONNECTOR_GLOB not in file_path
    }
    return {
        Connector(get_connector_name_from_path(changed_file))
        for changed_file in changed_source_connector_files
    }


def _get_relative_connector_folder_name_from_metadata_path(
    metadata_file_path: str,
) -> str:
    """Get the relative connector folder name from the metadata file path.

    Args:
        metadata_file_path (Path): Path to the metadata file.

    Returns:
        str: The relative connector folder name.
    """
    # remove CONNECTOR_PATH_PREFIX and anything before
    metadata_file_path = metadata_file_path.split(CONNECTOR_PATH_PREFIX)[-1]

    # remove metadata.yaml
    metadata_file_path = metadata_file_path.replace(METADATA_FILE_NAME, "")

    # remove leading and trailing slashes
    metadata_file_path = metadata_file_path.strip("/")
    return metadata_file_path


def get_all_connectors_in_repo() -> Set[Connector]:
    """Retrieve a set of all Connectors in the repo.
    We globe the connectors folder for metadata.yaml files and construct Connectors from the directory name.

    Returns:
        A set of Connectors.
    """
    repo = git.Repo(search_parent_directories=True)
    repo_path = repo.working_tree_dir

    return {
        Connector(_get_relative_connector_folder_name_from_metadata_path(metadata_file))
        for metadata_file in glob(
            f"{repo_path}/{CONNECTOR_PATH_PREFIX}/**/metadata.yaml", recursive=True
        )
        if SCAFFOLD_CONNECTOR_GLOB not in metadata_file
    }


class ConnectorTypeEnum(str, Enum):
    source = "source"
    destination = "destination"


class SupportLevelEnum(str, Enum):
    certified = "certified"
    community = "community"
    archived = "archived"



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Unit tests for the bump_deps module."""

from __future__ import annotations

import subprocess
import tempfile
from pathlib import Path
from unittest.mock import patch

import pytest

from airbyte_ops_mcp.airbyte_repo.bump_deps import (
    DepsError,
    _get_outdated_packages,
    bump_deps,
)


def _make_connector(
    tmpdir: str,
    connector_name: str,
    *,
    has_lock: bool = True,
) -> Path:
    """Create a minimal connector directory with pyproject.toml and optional lock file."""
    connector = Path(tmpdir) / "airbyte-integrations" / "connectors" / connector_name
    connector.mkdir(parents=True)
    (connector / "pyproject.toml").write_text(
        '[tool.poetry]\nname = "test"\nversion = "0.1.0"\n'
        '\n[tool.poetry.dependencies]\npython = "^3.9"\n'
        'airbyte-cdk = ">=7.0,<8.0"\nrequests = "*"\npydantic = "^2.0"\n'
    )
    if has_lock:
        (connector / "poetry.lock").write_text("# lock content v1\n")
    return connector


# ---------------------------------------------------------------------------
# Non-Poetry connectors are a no-op
# ---------------------------------------------------------------------------


@pytest.mark.unit
@patch("airbyte_ops_mcp.airbyte_repo.bump_deps._detect_connector_language")
def test_manifest_only_is_noop(mock_lang):
    mock_lang.return_value = "manifest-only"
    with tempfile.TemporaryDirectory() as tmpdir:
        connector = Path(tmpdir) / "airbyte-integrations" / "connectors" / "source-test"
        connector.mkdir(parents=True)
        result = bump_deps(tmpdir, "source-test")

    assert result.updated is False
    assert "does not use Poetry" in result.message


@pytest.mark.unit
@patch("airbyte_ops_mcp.airbyte_repo.bump_deps._detect_connector_language")
def test_java_is_noop(mock_lang):
    mock_lang.return_value = "java"
    with tempfile.TemporaryDirectory() as tmpdir:
        connector = Path(tmpdir) / "airbyte-integrations" / "connectors" / "source-test"
        connector.mkdir(parents=True)
        result = bump_deps(tmpdir, "source-test")

    assert result.updated is False


# ---------------------------------------------------------------------------
# Missing poetry.lock returns early
# ---------------------------------------------------------------------------


@pytest.mark.unit
@patch("airbyte_ops_mcp.airbyte_repo.bump_deps._detect_connector_language")
def test_no_lock_file_returns_early(mock_lang):
    mock_lang.return_value = "python"
    with tempfile.TemporaryDirectory() as tmpdir:
        _make_connector(tmpdir, "source-test", has_lock=False)
        result = bump_deps(tmpdir, "source-test")

    assert result.updated is False
    assert "No poetry.lock" in result.message


# ---------------------------------------------------------------------------
# Dry run reports correct files
# ---------------------------------------------------------------------------


@pytest.mark.unit
@patch("airbyte_ops_mcp.airbyte_repo.bump_deps._detect_connector_language")
def test_dry_run_reports_lock_file(mock_lang):
    mock_lang.return_value = "python"
    with tempfile.TemporaryDirectory() as tmpdir:
        _make_connector(tmpdir, "source-test")
        result = bump_deps(tmpdir, "source-test", dry_run=True)

    assert result.dry_run is True
    assert len(result.files_modified) == 1
    assert "poetry.lock" in result.files_modified[0]
    assert "source-test" in result.files_modified[0]


# ---------------------------------------------------------------------------
# _get_outdated_packages
# ---------------------------------------------------------------------------


OUTDATED_OUTPUT = (
    "airbyte-cdk                         (!) 7.10.1    7.15.0      \n"
    "pytest                              (!) 8.4.2     9.0.2       \n"
    "responses                           (!) 0.23.3    0.26.0      \n"
)


@pytest.mark.unit
@patch("airbyte_ops_mcp.airbyte_repo.bump_deps.subprocess.run")
def test_get_outdated_packages_parses_output(mock_run):
    mock_run.return_value = subprocess.CompletedProcess(
        args=[], returncode=0, stdout=OUTDATED_OUTPUT, stderr=""
    )
    result = _get_outdated_packages(Path("/fake/connector"), "source-test")
    assert result == ["airbyte-cdk", "pytest", "responses"]
    call_args = mock_run.call_args[0][0]
    assert call_args == ["poetry", "show", "--outdated", "--top-level", "--no-ansi"]


@pytest.mark.unit
@patch("airbyte_ops_mcp.airbyte_repo.bump_deps.subprocess.run")
def test_get_outdated_packages_empty(mock_run):
    mock_run.return_value = subprocess.CompletedProcess(
        args=[], returncode=0, stdout="", stderr=""
    )
    result = _get_outdated_packages(Path("/fake/connector"), "source-test")
    assert result == []


@pytest.mark.unit
@patch("airbyte_ops_mcp.airbyte_repo.bump_deps.subprocess.run")
def test_get_outdated_packages_failure_raises(mock_run):
    mock_run.side_effect = subprocess.CalledProcessError(1, "poetry", stderr="boom")
    with pytest.raises(DepsError, match="failed"):
        _get_outdated_packages(Path("/fake/connector"), "source-test")


# ---------------------------------------------------------------------------
# Actual update uses outdated-package check instead of file diff
# ---------------------------------------------------------------------------


@pytest.mark.unit
@patch("airbyte_ops_mcp.airbyte_repo.bump_deps._detect_connector_language")
@patch("airbyte_ops_mcp.airbyte_repo.bump_deps._get_outdated_packages")
@patch("airbyte_ops_mcp.airbyte_repo.bump_deps.subprocess.run")
def test_no_outdated_packages_skips_update(mock_run, mock_outdated, mock_lang):
    """When no packages are outdated, skip poetry update entirely -> updated=False."""
    mock_lang.return_value = "python"
    mock_outdated.return_value = []

    with tempfile.TemporaryDirectory() as tmpdir:
        _make_connector(tmpdir, "source-test")
        result = bump_deps(tmpdir, "source-test")

    assert result.updated is False
    assert result.outdated_packages == []
    assert "already up to date" in result.message
    # poetry update --lock should NOT have been called
    mock_run.assert_not_called()


@pytest.mark.unit
@patch("airbyte_ops_mcp.airbyte_repo.bump_deps._detect_connector_language")
@patch("airbyte_ops_mcp.airbyte_repo.bump_deps._get_outdated_packages")
@patch("airbyte_ops_mcp.airbyte_repo.bump_deps.subprocess.run")
def test_outdated_packages_triggers_update(mock_run, mock_outdated, mock_lang):
    """When packages are outdated, run poetry update and report updated=True."""
    mock_lang.return_value = "python"
    mock_outdated.return_value = ["airbyte-cdk", "requests"]
    mock_run.return_value = subprocess.CompletedProcess(args=[], returncode=0)

    with tempfile.TemporaryDirectory() as tmpdir:
        _make_connector(tmpdir, "source-test")
        result = bump_deps(tmpdir, "source-test")

    assert result.updated is True
    assert result.outdated_packages == ["airbyte-cdk", "requests"]
    assert len(result.files_modified) == 1
    # Verify poetry update --lock was called
    call_args = mock_run.call_args[0][0]
    assert call_args == ["poetry", "update", "--lock"]


# ---------------------------------------------------------------------------
# Subprocess failure raises DepsError
# ---------------------------------------------------------------------------


@pytest.mark.unit
@patch("airbyte_ops_mcp.airbyte_repo.bump_deps._detect_connector_language")
@patch("airbyte_ops_mcp.airbyte_repo.bump_deps._get_outdated_packages")
@patch("airbyte_ops_mcp.airbyte_repo.bump_deps.subprocess.run")
def test_poetry_update_failure_raises(mock_run, mock_outdated, mock_lang):
    mock_lang.return_value = "python"
    mock_outdated.return_value = ["airbyte-cdk"]
    mock_run.side_effect = subprocess.CalledProcessError(1, "poetry", stderr="boom")

    with tempfile.TemporaryDirectory() as tmpdir:
        _make_connector(tmpdir, "source-test")
        with pytest.raises(DepsError, match="failed"):
            bump_deps(tmpdir, "source-test")



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

from airbyte_ops_mcp._legacy.airbyte_ci.connector_qa.checks.documentation import (
    documentation,
)
from airbyte_ops_mcp._legacy.airbyte_ci.connector_qa.models import CheckStatus


class TestCheckMigrationGuide:
    def test_passed_when_no_breaking_changes(self, mocker):
        # Arrange
        connector = mocker.Mock(
            technical_name="test-connector", metadata={}, migration_guide_file_path=None
        )

        # Act
        result = documentation.CheckMigrationGuide()._run(connector)

        # Assert
        assert result.status == CheckStatus.PASSED
        assert (
            "No breaking changes found. A migration guide is not required"
            in result.message
        )

    def test_fail_when_migration_guide_file_path_does_not_exists(
        self, mocker, tmp_path
    ):
        # Arrange
        connector = mocker.Mock(
            technical_name="test-connector",
            metadata={"releases": {"breakingChanges": {"1.0.0": "Description"}}},
            migration_guide_file_path=tmp_path / "not_existing_migration_guide.md",
        )

        # Act
        result = documentation.CheckMigrationGuide()._run(connector)

        # Assert
        assert result.status == CheckStatus.FAILED
        assert "Migration guide file is missing " in result.message
        assert connector.technical_name in result.message
        assert "Please create a migration guide in" in result.message

    def test_fail_when_migration_guide_file_path_is_none(self, mocker):
        # Arrange
        connector = mocker.Mock(
            technical_name="test-connector",
            metadata={"releases": {"breakingChanges": {"1.0.0": "Description"}}},
            migration_guide_file_path=None,
        )

        # Act
        result = documentation.CheckMigrationGuide()._run(connector)

        # Assert
        assert result.status == CheckStatus.FAILED
        assert "Migration guide file is missing " in result.message
        assert connector.technical_name in result.message
        assert "Please create a migration guide in" in result.message

    def test_fail_when_migration_guide_file_does_not_contain_correct_header(
        self, mocker, tmp_path
    ):
        # Arrange
        connector = mocker.Mock(
            name_from_metadata="Test Connector",
            technical_name="test-connector",
            metadata={"releases": {"breakingChanges": {"1.0.0": "Description"}}},
            migration_guide_file_path=tmp_path / "migration_guide.md",
        )
        connector.migration_guide_file_path.write_text("")

        # Act
        result = documentation.CheckMigrationGuide()._run(connector)

        # Assert
        assert result.status == CheckStatus.FAILED
        assert (
            "Migration guide file for test-connector does not contain the expected header"
            in result.message
        )

    def test_pass_when_migration_guide_header_not_on_first_line(self, mocker, tmp_path):
        # Arrange
        connector = mocker.Mock(
            name_from_metadata="Test Connector",
            technical_name="test-connector",
            metadata={"releases": {"breakingChanges": {"1.0.0": "Description"}}},
            migration_guide_file_path=tmp_path / "migration_guide.md",
        )
        connector.migration_guide_file_path.write_text(
            "import MigrationGuide from '@site/static/_migration_guides_upgrade_guide.md';\n"
            "\n"
            "# Test Connector Migration Guide\n"
            "\n"
            "## Upgrading to 1.0.0\n"
        )

        # Act
        result = documentation.CheckMigrationGuide()._run(connector)

        # Assert
        assert result.status == CheckStatus.PASSED
        assert "The migration guide is correctly templated" in result.message

    def test_fail_when_migration_guide_file_has_missing_version_headings(
        self, mocker, tmp_path
    ):
        # Arrange
        connector = mocker.Mock(
            name_from_metadata="Test Connector",
            technical_name="test-connector",
            metadata={
                "releases": {
                    "breakingChanges": {"1.0.0": "Description", "2.0.0": "Description"}
                }
            },
            migration_guide_file_path=tmp_path / "migration_guide.md",
        )
        connector.migration_guide_file_path.write_text(
            "# Test Connector Migration Guide\n## Upgrading to 1.0.0\n"
        )

        # Act
        result = documentation.CheckMigrationGuide()._run(connector)

        # Assert
        assert result.status == CheckStatus.FAILED
        assert (
            f"Migration guide file for {connector.name_from_metadata} has incorrect version headings"
            in result.message
        )
        assert (
            "Expected headings: ['## Upgrading to 2.0.0', '## Upgrading to 1.0.0']"
            in result.message
        )

    def test_fail_when_migration_guide_file_has_invalid_version_headings(
        self, mocker, tmp_path
    ):
        # Arrange
        connector = mocker.Mock(
            name_from_metadata="Test Connector",
            technical_name="test-connector",
            metadata={
                "releases": {
                    "breakingChanges": {"1.0.0": "Description", "2.0.0": "Description"}
                }
            },
            migration_guide_file_path=tmp_path / "migration_guide.md",
        )
        connector.migration_guide_file_path.write_text(
            "# Test Connector Migration Guide\n## Upgrade to 1.0.0\n"
        )

        # Act
        result = documentation.CheckMigrationGuide()._run(connector)

        # Assert
        assert result.status == CheckStatus.FAILED
        assert (
            f"Migration guide file for {connector.name_from_metadata} has incorrect version headings"
            in result.message
        )
        assert (
            "Expected headings: ['## Upgrading to 2.0.0', '## Upgrading to 1.0.0']"
            in result.message
        )

    def test_fail_when_migration_guide_file_has_ascending_version_headings(
        self, mocker, tmp_path
    ):
        # Arrange
        connector = mocker.Mock(
            name_from_metadata="Test Connector",
            technical_name="test-connector",
            metadata={
                "releases": {
                    "breakingChanges": {"1.0.0": "Description", "2.0.0": "Description"}
                }
            },
            migration_guide_file_path=tmp_path / "migration_guide.md",
        )
        connector.migration_guide_file_path.write_text(
            "# Test Connector Migration Guide\n## Upgrading to 1.0.0\n## Upgrading to 2.0.0\n"
        )

        # Act
        result = documentation.CheckMigrationGuide()._run(connector)

        # Assert
        assert result.status == CheckStatus.FAILED
        assert (
            f"Migration guide file for {connector.name_from_metadata} has incorrect version headings"
            in result.message
        )
        assert (
            "Expected headings: ['## Upgrading to 2.0.0', '## Upgrading to 1.0.0']"
            in result.message
        )

    def test_fail_when_migration_guide_file_has_incorrect_version_headings(
        self, mocker, tmp_path
    ):
        # Arrange
        connector = mocker.Mock(
            name_from_metadata="Test Connector",
            technical_name="test-connector",
            metadata={
                "releases": {
                    "breakingChanges": {"1.0.0": "Description", "2.0.0": "Description"}
                }
            },
            migration_guide_file_path=tmp_path / "migration_guide.md",
        )
        connector.migration_guide_file_path.write_text(
            "# Test Connector Migration Guide\n## Upgrading to 1.0.0\n## Upgrading to 3.0.0\n"
        )

        # Act
        result = documentation.CheckMigrationGuide()._run(connector)

        # Assert
        assert result.status == CheckStatus.FAILED
        assert (
            f"Migration guide file for {connector.name_from_metadata} has incorrect version headings"
            in result.message
        )
        assert (
            "Expected headings: ['## Upgrading to 2.0.0', '## Upgrading to 1.0.0']"
            in result.message
        )

    def test_pass_when_migration_guide_file_has_correct_version_headings(
        self, mocker, tmp_path
    ):
        # Arrange
        connector = mocker.Mock(
            name_from_metadata="Test Connector",
            technical_name="test-connector",
            metadata={
                "releases": {
                    "breakingChanges": {"1.0.0": "Description", "2.0.0": "Description"}
                }
            },
            migration_guide_file_path=tmp_path / "migration_guide.md",
        )
        connector.migration_guide_file_path.write_text(
            "# Test Connector Migration Guide\n## Upgrading to 2.0.0\n## Upgrading to 1.0.0\n"
        )

        # Act
        result = documentation.CheckMigrationGuide()._run(connector)

        # Assert
        assert result.status == CheckStatus.PASSED
        assert "The migration guide is correctly templated" in result.message

    def test_pass_when_migration_guide_file_has_import_before_header(
        self, mocker, tmp_path
    ):
        # Arrange
        connector = mocker.Mock(
            name_from_metadata="Test Connector",
            technical_name="test-connector",
            metadata={
                "releases": {
                    "breakingChanges": {"1.0.0": "Description", "2.0.0": "Description"}
                }
            },
            migration_guide_file_path=tmp_path / "migration_guide.md",
        )
        connector.migration_guide_file_path.write_text(
            "import MigrationGuide from '@site/static/_migration_guides_upgrade_guide.md';\n"
            "\n"
            "# Test Connector Migration Guide\n"
            "\n"
            "## Upgrading to 2.0.0\n"
            "## Upgrading to 1.0.0\n"
        )

        # Act
        result = documentation.CheckMigrationGuide()._run(connector)

        # Assert
        assert result.status == CheckStatus.PASSED
        assert "The migration guide is correctly templated" in result.message


class TestCheckDocumentationExists:
    def test_fail_when_documentation_file_path_is_none(self, mocker):
        # Arrange
        connector = mocker.Mock(
            technical_name="test-connector", documentation_file_path=None
        )

        # Act
        result = documentation.CheckDocumentationExists()._run(connector)

        # Assert
        assert result.status == CheckStatus.FAILED
        assert (
            "User facing documentation file is missing. Please create it"
            in result.message
        )

    def test_fail_when_documentation_file_path_does_not_exists(self, mocker, tmp_path):
        # Arrange
        connector = mocker.Mock(
            technical_name="test-connector",
            documentation_file_path=tmp_path / "not_existing_documentation.md",
        )

        # Act
        result = documentation.CheckDocumentationExists()._run(connector)

        # Assert
        assert result.status == CheckStatus.FAILED
        assert (
            "User facing documentation file is missing. Please create it"
            in result.message
        )

    def test_pass_when_documentation_file_path_exists(self, mocker, tmp_path):
        # Arrange
        documentation_file_path = tmp_path / "documentation.md"
        connector = mocker.Mock(
            technical_name="test-connector",
            documentation_file_path=documentation_file_path,
        )
        connector.documentation_file_path.write_text("")

        # Act
        result = documentation.CheckDocumentationExists()._run(connector)

        # Assert
        assert result.status == CheckStatus.PASSED
        assert (
            f"User facing documentation file {documentation_file_path} exists"
            in result.message
        )


class TestCheckDocumentationContent:
    def test_fail_when_documentation_file_path_does_not_exists(self, mocker, tmp_path):
        # Arrange
        connector = mocker.Mock(
            technical_name="test-connector",
            ab_internal_sl=300,
            language="python",
            connector_type="source",
            documentation_file_path=tmp_path / "not_existing_documentation.md",
        )

        # Act
        result = documentation.CheckDocumentationHeadersOrder()._run(connector)

        # Assert
        assert result.status == CheckStatus.FAILED
        assert (
            "Could not check documentation structure as the documentation file is missing"
            in result.message
        )

    def test_fail_when_documentation_file_path_is_none(self, mocker):
        # Arrange
        connector = mocker.Mock(
            technical_name="test-connector",
            ab_internal_sl=300,
            language="python",
            connector_type="source",
            documentation_file_path=None,
        )

        # Act
        result = documentation.CheckDocumentationHeadersOrder()._run(connector)

        # Assert
        assert result.status == CheckStatus.FAILED
        assert (
            "Could not check documentation structure as the documentation file is missing"
            in result.message
        )

    def test_fail_when_documentation_file_is_empty(self, mocker, tmp_path):
        # Arrange
        documentation_file_path = tmp_path / "documentation.md"
        connector = mocker.Mock(
            technical_name="test-connector",
            documentation_file_path=documentation_file_path,
        )
        connector.documentation_file_path.write_text("")

        # Act
        result = documentation.CheckDocumentationHeadersOrder()._run(connector)

        # Assert
        assert result.status == CheckStatus.FAILED
        assert "Documentation file is empty" in result.message

    def test_fail_when_documentation_file_has_invalid_links(
        self, connector_with_invalid_links_in_documentation
    ):
        # Act
        result = documentation.CheckDocumentationLinks()._run(
            connector_with_invalid_links_in_documentation
        )

        # Assert
        assert result.status == CheckStatus.FAILED
        assert "Connector documentation uses invalid links:" in result.message
        assert "https://github.com/invalid-link with 404 status code" in result.message

    def test_fail_when_documentation_file_has_missing_headers(
        self, connector_with_invalid_documentation
    ):
        # Act
        result = documentation.CheckDocumentationHeadersOrder()._run(
            connector_with_invalid_documentation
        )

        # Assert
        assert result.status == CheckStatus.FAILED
        assert (
            "Documentation headers ordering/naming doesn't follow guidelines:"
            in result.message
        )
        assert (
            "Actual Heading: 'For Airbyte Cloud:'. Expected Heading: 'Setup guide'"
            in result.message
        )

    def test_fail_when_documentation_file_not_have_all_required_fields_in_prerequisites_section_content(
        self, connector_with_invalid_documentation
    ):
        # Act
        result = documentation.CheckPrerequisitesSectionDescribesRequiredFieldsFromSpec()._run(
            connector_with_invalid_documentation
        )

        # Assert
        assert result.status == CheckStatus.FAILED
        assert (
            "Missing descriptions for required spec fields: github repositories"
            in result.message
        )

    def test_fail_when_documentation_file_has_invalid_source_section_content(
        self, connector_with_invalid_documentation
    ):
        # Act
        result = documentation.CheckSourceSectionContent()._run(
            connector_with_invalid_documentation
        )

        # Assert
        assert result.status == CheckStatus.FAILED
        assert (
            "Connector GitHub section content does not follow standard template:"
            in result.message
        )
        assert (
            "+ This page contains the setup guide and reference information for the [GitHub]({docs_link}) source connector."
            in result.message
        )

    def test_success_when_documentation_file_has_invalid_for_airbyte_cloud_section_content(
        self, connector_with_invalid_documentation
    ):
        # We now allow for more lenient matching of the expected content. Leaving the old test here for reference.
        # Act
        result = documentation.CheckForAirbyteCloudSectionContent()._run(
            connector_with_invalid_documentation
        )

        # Assert
        assert result.status == CheckStatus.PASSED
        assert "Documentation guidelines are followed" in result.message
        # assert result.status == CheckStatus.FAILED
        # assert "Connector For Airbyte Cloud: section content does not follow standard template:" in result.message
        # assert "+ 1. [Log into your Airbyte Cloud](https://cloud.airbyte.com/workspaces) account." in result.message

    def test_fail_when_documentation_file_has_invalid_for_airbyte_open_section_content(
        self, connector_with_invalid_documentation
    ):
        # We now allow for more lenient matching of the expected content. Leaving the old test here for reference.
        # Act
        result = documentation.CheckForAirbyteOpenSectionContent()._run(
            connector_with_invalid_documentation
        )

        # Assert
        assert result.status == CheckStatus.PASSED
        assert "Documentation guidelines are followed" in result.message
        # assert result.status == CheckStatus.FAILED
        # assert "Connector For Airbyte Open Source: section content does not follow standard template" in result.message
        # assert "+ 1. Navigate to the Airbyte Open Source dashboard." in result.message

    def test_fail_when_documentation_file_has_invalid_supported_sync_modes_section_content(
        self, connector_with_invalid_documentation
    ):
        # We now allow for more lenient matching of the expected content. Leaving the old test here for reference.
        # Act
        result = documentation.CheckSupportedSyncModesSectionContent()._run(
            connector_with_invalid_documentation
        )

        # Assert
        assert result.status == CheckStatus.PASSED
        assert "Documentation guidelines are followed" in result.message

        # assert result.status == CheckStatus.FAILED
        # assert "Connector Supported sync modes section content does not follow standard template:" in result.message
        # assert (
        #     "+ The GitHub source connector supports the following"
        #     " [sync modes](https://docs.airbyte.com/cloud/core-concepts/#connection-sync-modes):"
        # ) in result.message

    def test_fail_when_documentation_file_has_invalid_tutorials_section_content(
        self, connector_with_invalid_documentation
    ):
        # We now allow for more lenient matching of the expected content. Leaving the old test here for reference.
        # Act
        result = documentation.CheckTutorialsSectionContent()._run(
            connector_with_invalid_documentation
        )

        # Assert
        assert result.status == CheckStatus.PASSED
        assert "Documentation guidelines are followed" in result.message

        # assert result.status == CheckStatus.FAILED
        # assert "Connector Tutorials section content does not follow standard template:" in result.message
        # assert "+ Now that you have set up the GitHub source connector, check out the following GitHub tutorials:" in result.message

    def test_fail_when_documentation_file_has_invalid_changelog_section_content(
        self, connector_with_invalid_documentation
    ):
        # Act
        result = documentation.CheckChangelogSectionContent()._run(
            connector_with_invalid_documentation
        )

        # Assert
        assert result.status == CheckStatus.FAILED
        assert (
            "Connector Changelog section content does not follow standard template:"
            in result.message
        )
        assert (
            "+ <details>\n+   <summary>Expand to review</summary>\n+ </details>"
            in result.message
        )

    def test_pass_when_documentation_file_has_correct_headers(
        self, connector_with_correct_documentation
    ):
        # Act
        result = documentation.CheckDocumentationHeadersOrder()._run(
            connector_with_correct_documentation
        )

        # Assert
        assert result.status == CheckStatus.PASSED
        assert result.message == "Documentation guidelines are followed"

    def test_pass_when_documentation_file_has_correct_prerequisites_section_content(
        self, connector_with_correct_documentation
    ):
        # Act
        result = documentation.CheckPrerequisitesSectionDescribesRequiredFieldsFromSpec()._run(
            connector_with_correct_documentation
        )

        # Assert
        assert result.status == CheckStatus.PASSED
        assert (
            "All required fields from spec are present in the connector documentation"
            in result.message
        )

    def test_pass_when_documentation_file_has_correct_source_section_content(
        self, connector_with_correct_documentation
    ):
        # Act
        result = documentation.CheckSourceSectionContent()._run(
            connector_with_correct_documentation
        )

        # Assert
        assert result.status == CheckStatus.PASSED
        assert "Documentation guidelines are followed" in result.message

    def test_pass_when_documentation_file_has_correct_for_airbyte_cloud_section_content(
        self, connector_with_correct_documentation
    ):
        # Act
        result = documentation.CheckForAirbyteCloudSectionContent()._run(
            connector_with_correct_documentation
        )

        # Assert
        assert result.status == CheckStatus.PASSED
        assert "Documentation guidelines are followed" in result.message

    def test_pass_when_documentation_file_has_correct_for_airbyte_open_section_content(
        self, connector_with_correct_documentation
    ):
        # Act
        result = documentation.CheckForAirbyteOpenSectionContent()._run(
            connector_with_correct_documentation
        )

        # Assert
        assert result.status == CheckStatus.PASSED
        assert "Documentation guidelines are followed" in result.message

    def test_pass_when_documentation_file_has_correct_supported_sync_modes_section_content(
        self, connector_with_correct_documentation
    ):
        # Act
        result = documentation.CheckSupportedSyncModesSectionContent()._run(
            connector_with_correct_documentation
        )

        # Assert
        assert result.status == CheckStatus.PASSED
        assert "Documentation guidelines are followed" in result.message

    def test_pass_when_documentation_file_has_correct_tutorials_section_content(
        self, connector_with_correct_documentation
    ):
        # Act
        result = documentation.CheckTutorialsSectionContent()._run(
            connector_with_correct_documentation
        )

        # Assert
        assert result.status == CheckStatus.PASSED
        assert "Documentation guidelines are followed" in result.message

    def test_pass_when_documentation_file_has_correct_headers_order(
        self, connector_with_correct_documentation
    ):
        # Act
        result = documentation.CheckDocumentationHeadersOrder()._run(
            connector_with_correct_documentation
        )

        # Assert
        assert result.status == CheckStatus.PASSED
        assert "Documentation guidelines are followed" in result.message

    def test_pass_when_documentation_file_has_correct_changelog_section_content(
        self, connector_with_correct_documentation
    ):
        # Act
        result = documentation.CheckChangelogSectionContent()._run(
            connector_with_correct_documentation
        )

        # Assert
        assert result.status == CheckStatus.PASSED
        assert "Documentation guidelines are followed" in result.message

    def test_pass_when_all_links_are_valid(self, connector_with_correct_documentation):
        # Act
        result = documentation.CheckDocumentationLinks()._run(
            connector_with_correct_documentation
        )

        # Assert
        assert result.status == CheckStatus.PASSED
        assert "Documentation links are valid" in result.message


class TestCheckChangelogEntry:
    def test_fail_when_documentation_file_path_does_not_exists(self, mocker, tmp_path):
        # Arrange
        connector = mocker.Mock(
            technical_name="test-connector",
            documentation_file_path=tmp_path / "not_existing_documentation.md",
        )

        # Act
        result = documentation.CheckChangelogEntry()._run(connector)

        # Assert
        assert result.status == CheckStatus.FAILED
        assert (
            "Could not check changelog entry as the documentation file is missing. Please create it."
            in result.message
        )

    def test_fail_when_documentation_file_path_is_none(self, mocker):
        # Arrange
        connector = mocker.Mock(
            technical_name="test-connector", documentation_file_path=None
        )

        # Act
        result = documentation.CheckChangelogEntry()._run(connector)

        # Assert
        assert result.status == CheckStatus.FAILED
        assert (
            "Could not check changelog entry as the documentation file is missing. Please create it."
            in result.message
        )

    def test_fail_when_documentation_file_is_empty(self, mocker, tmp_path):
        # Arrange
        documentation_file_path = tmp_path / "documentation.md"
        connector = mocker.Mock(
            technical_name="test-connector",
            documentation_file_path=documentation_file_path,
        )
        connector.documentation_file_path.write_text("")

        # Act
        result = documentation.CheckChangelogEntry()._run(connector)

        # Assert
        assert result.status == CheckStatus.FAILED
        assert "Documentation file is empty" in result.message

    def test_fail_when_documentation_file_doesnt_have_changelog_section(
        self, mocker, tmp_path
    ):
        # Arrange
        documentation_file_path = tmp_path / "documentation.md"
        connector = mocker.Mock(
            technical_name="test-connector",
            documentation_file_path=documentation_file_path,
        )
        connector.documentation_file_path.write_text("# Test Connector")

        # Act
        result = documentation.CheckChangelogEntry()._run(connector)

        # Assert
        assert result.status == CheckStatus.FAILED
        assert (
            "Connector documentation is missing a 'Changelog' section" in result.message
        )

    def test_fail_when_documentation_file_doesnt_have_changelog_entry(
        self, mocker, tmp_path
    ):
        # Arrange
        documentation_file_path = tmp_path / "documentation.md"
        connector = mocker.Mock(
            technical_name="test-connector",
            version="1.0.0",
            documentation_file_path=documentation_file_path,
        )
        connector.documentation_file_path.write_text("# Test Connector\n## Changelog")

        # Act
        result = documentation.CheckChangelogEntry()._run(connector)

        # Assert
        assert result.status == CheckStatus.FAILED
        assert (
            "Connectors must have a changelog entry for each version: changelog entry for version 1.0.0"
            in result.message
        )

    def test_pass_when_documentation_file_has_changelog_entry(self, mocker, tmp_path):
        # Arrange
        documentation_file_path = tmp_path / "documentation.md"
        connector = mocker.Mock(
            technical_name="test-connector",
            version="1.0.0",
            documentation_file_path=documentation_file_path,
        )
        connector.documentation_file_path.write_text(
            "# Test Connector\n## Changelog\n- Version 1.0.0"
        )

        # Act
        result = documentation.CheckChangelogEntry()._run(connector)

        # Assert
        assert result.status == CheckStatus.PASSED
        assert "Changelog entry found for version 1.0.0" in result.message



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Models for live tests - connector testing without Dagger."""

from __future__ import annotations

import contextlib
import logging
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum
from functools import cached_property
from pathlib import Path
from typing import Any, Iterator

from airbyte_protocol.models import (
    AirbyteCatalog,
    AirbyteMessage,
    ConfiguredAirbyteCatalog,
)
from airbyte_protocol.models import Type as AirbyteMessageType
from pydantic import ValidationError


class Command(Enum):
    """Airbyte connector commands."""

    CHECK = "check"
    DISCOVER = "discover"
    READ = "read"
    READ_WITH_STATE = "read-with-state"
    SPEC = "spec"

    def needs_config(self) -> bool:
        return self in {
            Command.CHECK,
            Command.DISCOVER,
            Command.READ,
            Command.READ_WITH_STATE,
        }

    def needs_catalog(self) -> bool:
        return self in {Command.READ, Command.READ_WITH_STATE}

    def needs_state(self) -> bool:
        return self in {Command.READ_WITH_STATE}


class TargetOrControl(Enum):
    """Indicates whether a connector is the target (new) or control (baseline) version."""

    TARGET = "target"
    CONTROL = "control"


class ActorType(Enum):
    """Type of Airbyte actor."""

    SOURCE = "source"
    DESTINATION = "destination"


@dataclass
class ConnectorUnderTest:
    """Represents a connector being tested.

    In validation tests, there would be one connector under test.
    When running regression tests, there would be two connectors under test:
    the target and the control versions of the same connector.
    """

    image_name: str
    target_or_control: TargetOrControl

    @property
    def name(self) -> str:
        """Get connector name without registry prefix."""
        return self.image_name.replace("airbyte/", "").split(":")[0]

    @property
    def name_without_type_prefix(self) -> str:
        """Get connector name without actor type prefix."""
        return self.name.replace(f"{self.actor_type.value}-", "")

    @property
    def version(self) -> str:
        """Get connector version from image tag."""
        return self.image_name.replace("airbyte/", "").split(":")[1]

    @property
    def actor_type(self) -> ActorType:
        """Infer actor type from image name."""
        if "airbyte/destination-" in self.image_name:
            return ActorType.DESTINATION
        elif "airbyte/source-" in self.image_name:
            return ActorType.SOURCE
        else:
            raise ValueError(
                f"Can't infer the actor type. Connector image name {self.image_name} "
                "does not contain 'airbyte/source' or 'airbyte/destination'"
            )

    @classmethod
    def from_image_name(
        cls,
        image_name: str,
        target_or_control: TargetOrControl,
    ) -> ConnectorUnderTest:
        """Create a ConnectorUnderTest from an image name."""
        return cls(image_name, target_or_control)


@dataclass
class ExecutionInputs:
    """Inputs for executing a connector command."""

    connector_under_test: ConnectorUnderTest
    command: Command
    output_dir: Path
    config: dict[str, Any] | None = None
    configured_catalog: ConfiguredAirbyteCatalog | None = None
    state: dict[str, Any] | None = None
    environment_variables: dict[str, str] | None = None

    def __post_init__(self) -> None:
        """Validate that required inputs are present for the command."""
        if self.command.needs_config() and self.config is None:
            raise ValueError(f"Config is required for {self.command.value} command")
        if self.command.needs_catalog() and self.configured_catalog is None:
            raise ValueError(f"Catalog is required for {self.command.value} command")
        if self.command.needs_state() and self.state is None:
            raise ValueError(f"State is required for {self.command.value} command")


@dataclass
class ExecutionResult:
    """Result of executing a connector command."""

    connector_under_test: ConnectorUnderTest
    command: Command
    stdout_file_path: Path
    stderr_file_path: Path
    success: bool
    exit_code: int
    configured_catalog: ConfiguredAirbyteCatalog | None = None
    config: dict[str, Any] | None = None
    _airbyte_messages: list[AirbyteMessage] = field(default_factory=list)
    _messages_loaded: bool = field(default=False, repr=False)

    @property
    def logger(self) -> logging.Logger:
        return logging.getLogger(
            f"{self.connector_under_test.target_or_control.value}-{self.command.value}"
        )

    @cached_property
    def airbyte_messages(self) -> list[AirbyteMessage]:
        """Parse and return all Airbyte messages from stdout."""
        if self._messages_loaded:
            return self._airbyte_messages

        messages = []
        for line in self.stdout_file_path.read_text().splitlines():
            line = line.strip()
            if not line:
                continue
            with contextlib.suppress(ValidationError):
                messages.append(AirbyteMessage.parse_raw(line))
        self._airbyte_messages = messages
        self._messages_loaded = True
        return messages

    @property
    def configured_streams(self) -> list[str]:
        """Get list of configured stream names."""
        if not self.configured_catalog:
            return []
        return [stream.stream.name for stream in self.configured_catalog.streams]

    def get_records(self) -> Iterator[AirbyteMessage]:
        """Iterate over record messages."""
        for message in self.airbyte_messages:
            if message.type is AirbyteMessageType.RECORD:
                yield message

    def get_records_per_stream(self, stream: str) -> Iterator[AirbyteMessage]:
        """Get records for a specific stream."""
        for message in self.get_records():
            if message.record.stream == stream:
                yield message

    def get_states(self) -> Iterator[AirbyteMessage]:
        """Iterate over state messages."""
        for message in self.airbyte_messages:
            if message.type is AirbyteMessageType.STATE:
                yield message

    def get_message_count_per_type(self) -> dict[AirbyteMessageType, int]:
        """Count messages by type."""
        counts: dict[AirbyteMessageType, int] = defaultdict(int)
        for message in self.airbyte_messages:
            counts[message.type] += 1
        return dict(counts)

    def get_record_count_per_stream(self) -> dict[str, int]:
        """Count records by stream name.

        Returns:
            Dictionary mapping stream names to record counts.
        """
        counts: dict[str, int] = defaultdict(int)
        for message in self.get_records():
            counts[message.record.stream] += 1
        return dict(counts)

    def get_catalog(self) -> AirbyteCatalog | None:
        """Get discovered catalog from messages."""
        for message in self.airbyte_messages:
            if message.type is AirbyteMessageType.CATALOG:
                return message.catalog
        return None

    def get_spec(self) -> Any | None:
        """Get connector spec from messages."""
        for message in self.airbyte_messages:
            if message.type is AirbyteMessageType.SPEC:
                return message.spec
        return None

    def get_connection_status(self) -> Any | None:
        """Get connection status from check command."""
        for message in self.airbyte_messages:
            if message.type is AirbyteMessageType.CONNECTION_STATUS:
                return message.connectionStatus
        return None

    def is_check_successful(self) -> bool:
        """Check if the check command was successful."""
        status = self.get_connection_status()
        if status is None:
            return False
        return status.status.value == "SUCCEEDED"

    def save_artifacts(self, output_dir: Path) -> None:
        """Save execution artifacts to the output directory."""
        output_dir.mkdir(parents=True, exist_ok=True)

        airbyte_messages_dir = output_dir / "airbyte_messages"
        airbyte_messages_dir.mkdir(parents=True, exist_ok=True)

        messages_by_type: dict[str, list[str]] = defaultdict(list)
        for message in self.airbyte_messages:
            type_name = message.type.value.lower()
            messages_by_type[type_name].append(message.model_dump_json())

        for type_name, messages in messages_by_type.items():
            file_path = airbyte_messages_dir / f"{type_name}.jsonl"
            file_path.write_text("\n".join(messages))

        # Save configured catalog (input) if available
        if self.configured_catalog is not None:
            catalog_path = output_dir / "configured_catalog.json"
            catalog_path.write_text(self.configured_catalog.model_dump_json(indent=2))
            self.logger.info(f"Saved configured catalog to {catalog_path}")

        self.logger.info(f"Artifacts saved to {output_dir}")



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Dependency bumping utilities for Airbyte connectors.

This module provides functionality to update all Poetry-managed dependencies
for connectors, refreshing the lock file to the latest versions allowed by
existing constraints.
"""

from __future__ import annotations

import subprocess
from dataclasses import dataclass, field
from pathlib import Path

from airbyte_ops_mcp.airbyte_repo.bump_version import (
    get_connector_path,
)
from airbyte_ops_mcp.airbyte_repo.list_connectors import (
    CONNECTOR_PATH_PREFIX,
    _detect_connector_language,
)

PYPROJECT_FILE_NAME = "pyproject.toml"
POETRY_LOCK_FILE_NAME = "poetry.lock"


class DepsError(Exception):
    """Base exception for dependency bump operations."""


@dataclass
class BumpDepsResult:
    """Result of a dependency bump operation."""

    connector: str
    language: str | None
    updated: bool
    dry_run: bool
    files_modified: list[str] = field(default_factory=list)
    message: str = ""
    outdated_packages: list[str] = field(default_factory=list)


def _get_outdated_packages(
    connector_path: Path,
    connector_name: str,
) -> list[str]:
    """Return a list of outdated top-level dependency names.

    Runs ``poetry show --outdated --top-level --no-ansi`` and parses the
    output.  Each line has the format::

        <package>  (!) <current>  <latest>  ...

    We return just the package names (first column).  An empty list means
    all dependencies are already at the latest allowed version.
    """
    try:
        proc = subprocess.run(
            ["poetry", "show", "--outdated", "--top-level", "--no-ansi"],
            cwd=str(connector_path),
            check=True,
            capture_output=True,
            text=True,
            timeout=300,
        )
    except FileNotFoundError as exc:
        raise DepsError(
            "Poetry is not installed or not on PATH. Cannot check outdated packages."
        ) from exc
    except subprocess.CalledProcessError as exc:
        raise DepsError(
            f"'poetry show --outdated' failed for connector '{connector_name}': "
            f"{exc.stderr or exc}"
        ) from exc
    except subprocess.TimeoutExpired as exc:
        raise DepsError(
            f"'poetry show --outdated' timed out for connector '{connector_name}'."
        ) from exc

    packages: list[str] = []
    for line in proc.stdout.strip().splitlines():
        parts = line.split()
        if parts:
            packages.append(parts[0])
    return packages


def bump_deps(
    repo_path: str | Path,
    connector_name: str,
    dry_run: bool = False,
) -> BumpDepsResult:
    """Update all dependencies for a connector.

    For Python / low-code connectors using Poetry, this runs
    `poetry update --lock` to refresh the lock file with the latest
    versions allowed by existing constraints in `pyproject.toml`.

    For connectors that do not use Poetry (manifest-only, Java, etc.),
    this is a no-op.

    Args:
        repo_path: Path to the Airbyte monorepo.
        connector_name: Technical name of the connector (e.g. `source-github`).
        dry_run: If `True`, report what would change without modifying files.

    Returns:
        A `BumpDepsResult` describing the outcome.

    Raises:
        ConnectorNotFoundError: If the connector directory does not exist.
        DepsError: On update failures.
    """
    repo_path = Path(repo_path)
    connector_path = get_connector_path(repo_path, connector_name)

    language = _detect_connector_language(connector_path, connector_name)

    # Only Poetry-based connectors have deps to update
    if language not in ("python", "low-code"):
        return BumpDepsResult(
            connector=connector_name,
            language=language,
            updated=False,
            dry_run=dry_run,
            message=(
                f"Connector '{connector_name}' (language={language}) does not use "
                f"Poetry. No dependencies to update."
            ),
        )

    pyproject_path = connector_path / PYPROJECT_FILE_NAME
    if not pyproject_path.exists():
        return BumpDepsResult(
            connector=connector_name,
            language=language,
            updated=False,
            dry_run=dry_run,
            message=f"No {PYPROJECT_FILE_NAME} found for connector '{connector_name}'.",
        )

    poetry_lock_path = connector_path / POETRY_LOCK_FILE_NAME
    if not poetry_lock_path.exists():
        return BumpDepsResult(
            connector=connector_name,
            language=language,
            updated=False,
            dry_run=dry_run,
            message=f"No {POETRY_LOCK_FILE_NAME} found for connector '{connector_name}'.",
        )

    if dry_run:
        return BumpDepsResult(
            connector=connector_name,
            language=language,
            updated=False,
            dry_run=True,
            files_modified=[
                f"{CONNECTOR_PATH_PREFIX}/{connector_name}/{POETRY_LOCK_FILE_NAME}"
            ],
            message="Dry run: would run 'poetry update --lock' to update dependencies.",
        )

    # Check for outdated packages before running the update.
    # This avoids false positives from Poetry reformatting the lock file
    # without actually changing any dependency versions.
    outdated_packages = _get_outdated_packages(connector_path, connector_name)

    if not outdated_packages:
        return BumpDepsResult(
            connector=connector_name,
            language=language,
            updated=False,
            dry_run=False,
            outdated_packages=[],
            message="Dependencies are already up to date.",
        )

    try:
        subprocess.run(
            ["poetry", "update", "--lock"],
            cwd=str(connector_path),
            check=True,
            capture_output=True,
            text=True,
            timeout=300,
        )
    except FileNotFoundError as exc:
        raise DepsError(
            "Poetry is not installed or not on PATH. Cannot update dependencies."
        ) from exc
    except subprocess.CalledProcessError as exc:
        raise DepsError(
            f"'poetry update --lock' failed for connector '{connector_name}': "
            f"{exc.stderr or exc}"
        ) from exc
    except subprocess.TimeoutExpired as exc:
        raise DepsError(
            f"'poetry update --lock' timed out for connector '{connector_name}'."
        ) from exc

    return BumpDepsResult(
        connector=connector_name,
        language=language,
        updated=True,
        dry_run=False,
        files_modified=[
            f"{CONNECTOR_PATH_PREFIX}/{connector_name}/{POETRY_LOCK_FILE_NAME}"
        ],
        outdated_packages=outdated_packages,
        message="Updated dependencies via 'poetry update --lock'.",
    )



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from consts import (
    CONNECTOR_TESTING_REQUIREMENTS,
    PIP_CACHE_PATH,
    PIP_CACHE_VOLUME_NAME,
    POETRY_CACHE_PATH,
    POETRY_CACHE_VOLUME_NAME,
    PYPROJECT_TOML_FILE_PATH,
)
from dagger import CacheSharingMode, CacheVolume, Client, Container
from pipelines.airbyte_ci.connectors.context import PipelineContext
from pipelines.helpers.utils import sh_dash_c


def with_python_base(
    context: PipelineContext, python_version: str = "3.11"
) -> Container:
    """Build a Python container with a cache volume for pip cache.

    Args:
        context (PipelineContext): The current test context, providing a dagger client and a repository directory.
        python_image_name (str, optional): The python image to use to build the python base environment. Defaults to "python:3.9-slim".

    Raises:
        ValueError: Raised if the python_image_name is not a python image.

    Returns:
        Container: The python base environment container.
    """

    pip_cache: CacheVolume = context.dagger_client.cache_volume("pip_cache")

    base_container = (
        context.dagger_client.container()
        .from_(f"python:{python_version}-slim")
        .with_mounted_cache("/root/.cache/pip", pip_cache)
        .with_exec(
            sh_dash_c(
                [
                    "apt-get update",
                    "apt-get install -y build-essential cmake g++ libffi-dev libstdc++6 git",
                    "pip install pip==23.1.2",
                ]
            )
        )
    )

    return base_container


def with_testing_dependencies(context: PipelineContext) -> Container:
    """Build a testing environment by installing testing dependencies on top of a python base environment.

    Args:
        context (PipelineContext): The current test context, providing a dagger client and a repository directory.

    Returns:
        Container: The testing environment container.
    """
    python_environment: Container = with_python_base(context)
    pyproject_toml_file = context.get_repo_dir(
        ".", include=[PYPROJECT_TOML_FILE_PATH]
    ).file(PYPROJECT_TOML_FILE_PATH)

    return python_environment.with_exec(
        ["pip", "install"] + CONNECTOR_TESTING_REQUIREMENTS
    ).with_file(f"/{PYPROJECT_TOML_FILE_PATH}", pyproject_toml_file)


def with_pip_cache(
    container: Container, dagger_client: Client, owner: str | None = None
) -> Container:
    """Mounts the pip cache in the container.
    Args:
        container (Container): A container with python installed
        owner (str, optional): The owner of the cache. Defaults to None.
    Returns:
        Container: A container with the pip cache mounted.
    """
    pip_cache_volume = dagger_client.cache_volume(PIP_CACHE_VOLUME_NAME)
    return container.with_mounted_cache(
        PIP_CACHE_PATH, pip_cache_volume, sharing=CacheSharingMode.SHARED, owner=owner
    )


def with_poetry_cache(
    container: Container, dagger_client: Client, owner: str | None = None
) -> Container:
    """Mounts the poetry cache in the container.
    Args:
        container (Container): A container with python installed
        owner (str, optional): The owner of the cache. Defaults to None.
    Returns:
        Container: A container with the poetry cache mounted.
    """
    poetry_cache_volume = dagger_client.cache_volume(POETRY_CACHE_VOLUME_NAME)
    return container.with_mounted_cache(
        POETRY_CACHE_PATH,
        poetry_cache_volume,
        sharing=CacheSharingMode.SHARED,
        owner=owner,
    )



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

"""This module contains hacks used in connectors pipelines. They're gathered here for tech debt visibility."""

from __future__ import annotations

from logging import Logger
from typing import TYPE_CHECKING, Callable, List

import asyncclick as click
from connector_ops.utils import ConnectorLanguage  # type: ignore
from pipelines import consts
from pipelines.airbyte_ci.steps.base_image import UpdateBaseImageMetadata
from pipelines.helpers.github import (
    AIRBYTE_GITHUB_REPO_URL,
    is_automerge_pull_request,
    update_commit_status_check,
)

if TYPE_CHECKING:
    from dagger import Container
    from pipelines.airbyte_ci.connectors.context import ConnectorContext
    from pipelines.models.steps import StepResult


async def cache_latest_cdk(context: ConnectorContext) -> None:
    """
    Download the latest CDK version to update the pip cache.

    Underlying issue:
        Most Python connectors, or normalization, are not pinning the CDK version they use.
        It means that the will get whatever version is in the pip cache.
        But the original goal of not pinning the CDK version is to always get the latest version.

    Hack:
        Call this function before building connector test environment to update the cache with the latest CDK version.

    Github Issue:
        Revisiting and aligning how we build Python connectors and using the same container for test, build and publish will provide better control over the CDK version.
        https://github.com/airbytehq/airbyte/issues/25523
    Args:
        dagger_client (Client): Dagger client.
    """
    # We want the CDK to be re-downloaded on every run per connector to ensure we always get the latest version.
    # But we don't want to invalidate the pip cache on every run because it could lead to a different CDK version installed on different architecture build.
    cachebuster_value = (
        f"{context.connector.technical_name}_{context.pipeline_start_timestamp}"
    )

    await (
        context.dagger_client.container()
        .from_("python:3.9-slim")
        .with_mounted_cache(
            consts.PIP_CACHE_PATH,
            context.dagger_client.cache_volume(consts.PIP_CACHE_VOLUME_NAME),
        )
        .with_env_variable("CACHEBUSTER", cachebuster_value)
        .with_exec(
            ["pip", "install", "--force-reinstall", "airbyte-cdk", "-vvv"],
            use_entrypoint=True,
        )
        .sync()
    )


def never_fail_exec(command: List[str]) -> Callable[[Container], Container]:
    """
    Wrap a command execution with some bash sugar to always exit with a 0 exit code but write the actual exit code to a file.

    Underlying issue:
        When a classic dagger with_exec is returning a >0 exit code an ExecError is raised.
        It's OK for the majority of our container interaction.
        But some execution, like running CAT, are expected to often fail.
        In CAT we don't want ExecError to be raised on container interaction because CAT might write updated secrets that we need to pull from the container after the test run.
        The bash trick below is a hack to always return a 0 exit code but write the actual exit code to a file.
        The file is then read by the pipeline to determine the exit code of the container.

    Args:
        command (List[str]): The command to run in the container.

    Returns:
        Callable: _description_
    """

    def never_fail_exec_inner(container: Container) -> Container:
        return container.with_exec(
            ["sh", "-c", f"{' '.join(command)}; echo $? > /exit_code"]
        )

    return never_fail_exec_inner


def do_regression_test_status_check(
    ctx: click.Context, status_check_name: str, logger: Logger
) -> None:
    """
    Emit a failing status check that requires a manual override, via a /-command.

    Only required for certified connectors.
    """
    commit = ctx.obj["git_revision"]
    run_url = ctx.obj["gha_workflow_run_url"]
    should_send = ctx.obj.get("ci_context") == consts.CIContext.PULL_REQUEST

    if (
        (not is_automerge_pull_request(ctx.obj.get("pull_request")))
        and (ctx.obj["git_repo_url"] == AIRBYTE_GITHUB_REPO_URL)
        and any(
            [
                (
                    connector.language == ConnectorLanguage.PYTHON
                    and connector.support_level == "certified"
                )
                for connector in ctx.obj["selected_connectors_with_modified_files"]
            ]
        )
    ):
        logger.info(
            f"is_automerge_pull_request={is_automerge_pull_request(ctx.obj.get('pull_request'))}"
        )
        logger.info(f"git_repo_url={ctx.obj['git_repo_url']}")
        for connector in ctx.obj["selected_connectors_with_modified_files"]:
            logger.info(f"connector = {connector.name}")
            logger.info(f"connector.language={connector.language}")
            logger.info(f"connector.support_level = {connector.support_level}")
        update_commit_status_check(
            commit,
            "failure",
            run_url,
            description="Check if regression tests have been manually approved",
            context=status_check_name,
            is_optional=False,
            should_send=should_send,
            logger=logger,
        )
    else:
        update_commit_status_check(
            commit,
            "success",
            run_url,
            description="[Skipped]",
            context=status_check_name,
            is_optional=True,
            should_send=should_send,
            logger=logger,
        )


def determine_changelog_entry_comment(
    upgrade_base_image_in_metadata_result: StepResult, default_comment: str
) -> str:
    assert isinstance(
        upgrade_base_image_in_metadata_result.step, UpdateBaseImageMetadata
    ), "StepResult's step must be instance of UpdateBaseImageMetadata"
    if (
        upgrade_base_image_in_metadata_result.output is not None
        and upgrade_base_image_in_metadata_result.output.get(
            "updated_base_image_address"
        )
    ):
        updated_base_image_address = upgrade_base_image_in_metadata_result.output.get(
            "updated_base_image_address"
        )
        if (
            "airbyte/python-connector-base:3.0.0" in updated_base_image_address
            or "airbyte/source-declarative-manifest:6.9.2" in updated_base_image_address
        ):
            return "Starting with this version, the Docker image is now rootless. Please note that this and future versions will not be compatible with Airbyte versions earlier than 0.64"
    return default_comment



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

import os
from typing import Dict, List, Set

from pydantic import BaseModel, Field, validator


class AirbyteCiPackageConfiguration(BaseModel):
    poe_tasks: Set[str] = Field(..., description="List of unique poe tasks to run")
    required_environment_variables: Set[str] = Field(
        set(),
        description="List of unique required environment variables to pass to the container running the poe task",
    )
    poetry_extras: Set[str] = Field(
        set(), description="List of unique poetry extras to install"
    )
    optional_poetry_groups: Set[str] = Field(
        set(), description="List of unique poetry groups to install"
    )
    side_car_docker_engine: bool = Field(
        False,
        description="Flag indicating the use of a sidecar Docker engine during the poe task executions",
    )
    mount_docker_socket: bool = Field(
        False,
        description="Flag indicating the mount of the host docker socket to the container running the poe task, useful when the package under test is using dagger",
    )
    python_versions: List[str] = Field(
        description="List of unique python versions to run the poe tasks on"
    )

    @validator("required_environment_variables")
    def check_required_environment_variables_are_set(cls, value: Set) -> Set:
        for required_env_var in value:
            if required_env_var not in os.environ:
                raise ValueError(f"Environment variable {required_env_var} is not set.")
        return value


def deserialize_airbyte_ci_config(
    pyproject_toml: Dict,
) -> AirbyteCiPackageConfiguration:
    try:
        airbyte_ci_config = pyproject_toml["tool"]["airbyte_ci"]
    except KeyError:
        raise ValueError("Missing tool.airbyte_ci configuration in pyproject.toml")
    return AirbyteCiPackageConfiguration.parse_obj(airbyte_ci_config)
